import mdtraj as md
import unittest

import mdciao.contacts
from mdciao.examples import filenames as test_filenames
from mdciao.utils import str_and_dict

#see https://stackoverflow.com/questions/169070/how-do-i-write-a-decorator-that-restores-the-cwd
import contextlib
@contextlib.contextmanager
def remember_cwd():
    curdir = os.getcwd()
    try:
        yield
    finally:
        os.chdir(curdir)

import shutil

from tempfile import TemporaryDirectory as _TDir

import os

from matplotlib import \
    pyplot as _plt

from mdciao import cli
from mdciao import fragments as mdcfragments
from mdciao import contacts

from mdciao.cli import *
#    residue_neighborhoods, \
#    sites, \
#    interface

from mdciao.nomenclature import \
    LabelerGPCR

from mdciao.parsers import \
    parser_for_CGN_overview, \
    parser_for_GPCR_overview

#from mdciao.contact_matrix_utils import \
#    contact_map

from tempfile import TemporaryDirectory

from unittest import mock
from pandas import \
    unique as _pandasunique

from os import \
    path as _path

import numpy as _np

class TestCLTBaseClass(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.geom = md.load(test_filenames.top_pdb)
        cls.traj = md.load(test_filenames.traj_xtc_stride_20, top=cls.geom.top)
        cls.traj_reverse = md.load(test_filenames.traj_xtc_stride_20, top=cls.geom.top)[::-1]

class Test_manage_timdep_plot_options(TestCLTBaseClass):

    @classmethod
    def setUpClass(cls):
        super(Test_manage_timdep_plot_options, cls).setUpClass()
        ctc_idxs = [[353, 348],
                    [353, 972],
                    [353, 347]]
        CPs = [contacts.ContactPair(pair,
                                    [md.compute_contacts(itraj, [pair])[0].squeeze() for itraj in
                                     [cls.traj,
                                      cls.traj_reverse[:10]]],
                                    [cls.traj.time,
                                     cls.traj_reverse.time[:10]],
                                    top=cls.geom.top,
                                    anchor_residue_idx=353)
               for pair in ctc_idxs]

        cls.ctc_grp = contacts.ContactGroup(CPs,
                                            top=cls.geom.top,
                                            neighbors_excluded=0,
                                            )

    """
    1:     0.55   LEU394-ARG389       0-0         353-348        33     0.55
    2:     0.47   LEU394-LYS270       0-3         353-972        71     1.02
    3:     0.38   LEU394-LEU388       0-0         353-347        32     1.39
    4:     0.23   LEU394-LEU230       0-3         353-957        56     1.62
    5:     0.10   LEU394-ARG385       0-0         353-344        29     1.73

    """

    def test_works(self):
        with _TDir(suffix="_test_mdciao") as tmpdir:
            myfig = self.ctc_grp.plot_timedep_ctcs(ctc_cutoff_Ang=3)
            fn= str_and_dict.FilenameGenerator("test_neig",3,tmpdir,"png","dat",150,"ps")
            cli._manage_timedep_ploting_and_saving_options(self.ctc_grp, fn, myfig,
                                                           savetrajs=True)
        _plt.close("all")

    def test_separate_N_ctcs(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            myfig = self.ctc_grp.plot_timedep_ctcs(pop_N_ctcs=True, ctc_cutoff_Ang=3)
            fn= str_and_dict.FilenameGenerator("test_neig",3,tmpdir,"png","dat",150,"ps")
            cli._manage_timedep_ploting_and_saving_options(self.ctc_grp,fn,myfig,
                                                           separate_N_ctcs=True,
                                                           savetrajs=True,
                                                           )
            _plt.close("all")
    def test_just_N_ctcs(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            myfig = self.ctc_grp.plot_timedep_ctcs(pop_N_ctcs=True, skip_timedep=True, ctc_cutoff_Ang=3)
            fn= str_and_dict.FilenameGenerator("test_neig",3,tmpdir,"png","dat",150,"ps")
            cli._manage_timedep_ploting_and_saving_options(self.ctc_grp, fn, myfig,
                                                           separate_N_ctcs=True,
                                                           plot_timedep=False,
                                                           )
            _plt.close("all")

    def test_no_files(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            fn = str_and_dict.FilenameGenerator("test_neig", 3, tmpdir, "png", "dat", 150, "ps")
            cli._manage_timedep_ploting_and_saving_options(self.ctc_grp, fn, [],
                                                           separate_N_ctcs=True,
                                                           plot_timedep=False,
                                                           )
            _plt.close("all")


class Test_residue_neighborhood(TestCLTBaseClass):

    @classmethod
    def setUpClass(cls):
        super(Test_residue_neighborhood, cls).setUpClass()
        cls.no_disk = True  # doesn't seem to speed things up much

    @mock.patch('builtins.input', lambda *args: '4')
    def test_neighborhoods_no_disk_works(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            with remember_cwd():
                os.chdir(tmpdir)
                input_values = (val for val in ["1.0"])
                with mock.patch('builtins.input', lambda *x: next(input_values)):
                    cli.residue_neighborhoods("200,395",
                                              self.traj,
                                              self.geom,
                                              no_disk=True
                                              )
                assert len(os.listdir(".")) == 0

    @mock.patch('builtins.input', lambda *args: '4')
    def test_neighborhoods(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            input_values = (val for val in ["1.0"])
            with mock.patch('builtins.input', lambda *x: next(input_values)):
                cli.residue_neighborhoods("200,395",
                                          [self.traj, self.traj_reverse],
                                          self.geom,
                                          output_dir=tmpdir,
                                          no_disk=self.no_disk)

    def test_no_top(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            input_values = (val for val in ["1.0"])
            with mock.patch('builtins.input', lambda *x: next(input_values)):
                cli.residue_neighborhoods("200,395",
                                          self.geom,
                                          output_dir=tmpdir,
                                          no_disk=self.no_disk)

    def test_res_idxs(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.residue_neighborhoods("1043",
                                      [self.traj, self.traj_reverse],
                                      self.geom,
                                      fragment_colors=True,
                                      allow_same_fragment_ctcs=False,
                                      short_AA_names=True,
                                      res_idxs=True,
                                      output_dir=tmpdir,
                                      no_disk=self.no_disk)

    def test_excel(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.residue_neighborhoods("395",
                                      [self.traj, self.traj_reverse],
                                      self.geom,
                                      table_ext=".xlsx",
                                      output_dir=tmpdir)

    def test_distro(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.residue_neighborhoods("395",
                                      [self.traj, self.traj_reverse],
                                      self.geom,
                                      distro=True,
                                      output_dir=tmpdir,
                                      no_disk=self.no_disk)

    def test_AAs(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.residue_neighborhoods("395",
                                      [self.traj, self.traj_reverse],
                                      self.geom,
                                      short_AA_names=True,
                                      output_dir=tmpdir,
                                      no_disk=self.no_disk)

    def test_colors(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.residue_neighborhoods("395",
                                      [self.traj, self.traj_reverse],
                                      self.geom,
                                      short_AA_names=True,
                                      fragment_colors=True,
                                      output_dir=tmpdir,
                                      no_disk=self.no_disk)

        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.residue_neighborhoods("395",
                                      [self.traj, self.traj_reverse],
                                      self.geom,
                                      short_AA_names=True,
                                      fragment_colors='c',
                                      output_dir=tmpdir,
                                      no_disk=self.no_disk)

    def test_no_residues_returns_None(self):
        # This is more of an argparse fail
        # TODO consider throwing exception
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            assert None == cli.residue_neighborhoods(None,
                                                     [self.traj, self.traj_reverse],
                                                     self.geom,
                                                     None,
                                                     distro=True,
                                                     output_dir=tmpdir,
                                                     no_disk=self.no_disk)

    def test_wrong_input_resSeq_idxs(self):
        with self.assertRaises(ValueError):
            with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
                cli.residue_neighborhoods("AX*",
                                          [self.traj, self.traj_reverse],
                                          self.geom,
                                          distro=True,
                                          output_dir=tmpdir,
                                          no_disk=self.no_disk)

    def test_nomenclature_GPCR(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.residue_neighborhoods("R131",
                                      [self.traj, self.traj_reverse],
                                      self.geom,
                                      GPCR_UniProt=test_filenames.adrb2_human_xlsx,
                                      output_dir=tmpdir,
                                      accept_guess=True,
                                      no_disk=self.no_disk
                                      )

    def test_nomenclature_CGN(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            with remember_cwd():
                os.chdir(tmpdir)
                cli.residue_neighborhoods("R131",
                                          [self.traj, self.traj_reverse],
                                          self.geom,
                                          CGN_UniProt="gnas2_human",
                                          output_dir=tmpdir,
                                          accept_guess=True,
                                          no_disk=self.no_disk
                                          )

    def test_nomenclature_CGN_and_GPCR(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            with remember_cwd():
                os.chdir(tmpdir)
                cli.residue_neighborhoods("R131",
                                          [self.traj, self.traj_reverse],
                                          self.geom,
                                          CGN_UniProt="gnas2_human",
                                          GPCR_UniProt=test_filenames.adrb2_human_xlsx,
                                          output_dir=tmpdir,
                                          accept_guess=True,
                                          no_disk=self.no_disk
                                          )

    def test_no_contacts_at_allp(self):
        cli.residue_neighborhoods("R131",
                                  [self.traj, self.traj_reverse],
                                  self.geom,
                                  ctc_cutoff_Ang=.1,
                                  )

    def test_some_CG_have_no_contacts(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.residue_neighborhoods("0-3",
                                      [self.traj, self.traj_reverse],
                                      self.geom,
                                      ctc_cutoff_Ang=3.2,
                                      res_idxs=True,
                                      output_dir=tmpdir,
                                      no_disk=self.no_disk
                                      )

    def test_no_bonds_fail(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            top = self.geom[0]
            top.top._bonds = []
            with self.assertRaises(ValueError):
                cli.residue_neighborhoods("R131",
                                          [self.traj, self.traj_reverse],
                                          top,
                                          output_dir=tmpdir)
    def test_naive_bonds_CAs(self):
        geom_CAs = self.geom[0].atom_slice(self.geom[0].top.select("name CA"))
        print(geom_CAs)
        with self.assertRaises(ValueError):
            cli.residue_neighborhoods("R131",
                                      [geom_CAs],
                                      no_disk=True,
                                      figures=False
                                      )
        n = cli.residue_neighborhoods("R131",
                                      [geom_CAs],
                                      no_disk=True,
                                      figures=False,
                                      naive_bonds=True,
                                      ctc_cutoff_Ang=1000,
                                      ctc_control=1.,
                                      )[861]
        assert n.n_ctcs==geom_CAs.top.n_residues-4-4-1# all residues minus 4 neighbors (x2 directions) minus itself

class Test_sites(TestCLTBaseClass):

    def test_sites(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
             cli.sites([test_filenames.tip_json],[self.traj, self.traj_reverse], self.geom,
                  output_dir=tmpdir)

    def test_sites_no_distk(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
             cli.sites([test_filenames.tip_json],[self.traj, self.traj_reverse], self.geom,
                  no_disk=True)

    def test_scheme_CA(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
             cli.sites([test_filenames.tip_json], [self.traj, self.traj_reverse],self.geom,
                  output_dir=tmpdir,
                  scheme="COM")

    def test_w_table_xlsx(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
             cli.sites([test_filenames.tip_json],[self.traj, self.traj_reverse], self.geom,
                  output_dir=tmpdir,
                  scheme="COM",
                       table_ext=".xlsx")

    def test_w_table_dat(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
             cli.sites([test_filenames.tip_json], [self.traj, self.traj_reverse],self.geom,
                  output_dir=tmpdir,
                  scheme="COM",
                       table_ext=".dat")
    def test_sites_no_top(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
             cli.sites([test_filenames.tip_json], [self.traj, self.traj_reverse],
                       output_dir=tmpdir)

    def test_sites_distro(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
             cli.sites([test_filenames.tip_json],
                       [self.traj, self.traj_reverse],
                       self.geom,
                       distro=True,
                       output_dir=tmpdir)

    def test_sites_w_incomplete_sites(self):
        # Only the last site will survive
        site_list = [
            {'name': 'site0', 'pairs': {'AAresSeq': ['ALA20-ALA21',  # ALA20 doesn't exist
                                                     'GLU101-GLU122']}},  # both exist

            {'name': 'site1', 'pairs': {'AAresSeq': ['GLU31-ALA20',  # GLU31 and ALA20 don't exist
                                                     'GLU17-GLU12']}},  # both exist

            {'name': 'site2', 'pairs': {'AAresSeq': ['GLN101-ALA122']}},  # GLN101 doesn't exist

            {'name': 'site3', 'pairs': {'AAresSeq': ['GLU101-GLU122']}}  # both exist, but was seen before
        ]
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            output_sites = cli.sites(site_list,
                      [self.traj, self.traj_reverse],
                      self.geom,
                      distro=True,
                      figures=False,
                      no_disk=True,
                      output_dir=tmpdir)
            assert len(output_sites)==1
            assert "site3" in output_sites.keys()
            assert output_sites["site3"]
            assert output_sites["site3"].n_ctcs == 1
            _np.testing.assert_array_equal(output_sites["site3"].res_idxs_pairs[0], [ 69, 852])

    def test_sites_w_incomplete_sites_allow_partial_sites(self):
        # Only the last site will survive
        site_list = [
            {'name': 'site0', 'pairs': {'AAresSeq': ['ALA20-ALA21',  # ALA20 doesn't exist
                                                     'GLU101-GLU122']}},  # both exist

            {'name': 'site1', 'pairs': {'AAresSeq': ['GLU31-ALA20',  # GLU31 and ALA20 don't exist
                                                     'GLU17-GLU12']}},  # both exist

            {'name': 'site2', 'pairs': {'AAresSeq': ['GLN101-ALA122']}},  # GLN101 doesn't exist

            {'name': 'site3', 'pairs': {'AAresSeq': ['GLU101-GLU122']}}  # both exist, but was seen before
        ]
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            output_sites = cli.sites(site_list,
                                     [self.traj, self.traj_reverse],
                                     self.geom,
                                     distro=True,
                                     figures=False,
                                     no_disk=True,
                                     output_dir=tmpdir,
                                     allow_partial_sites=True)
            assert len(output_sites)==3
            assert output_sites["site0"]
            assert output_sites["site0"].n_ctcs == 1
            _np.testing.assert_array_equal(output_sites["site0"].res_idxs_pairs[0], [69, 852])

            assert output_sites["site1"]
            assert output_sites["site1"].n_ctcs == 1
            _np.testing.assert_array_equal(output_sites["site1"].res_idxs_pairs[0], [709, 365])

            assert "site3" in output_sites.keys()
            assert output_sites["site3"]
            assert output_sites["site3"].n_ctcs == 1
            _np.testing.assert_array_equal(output_sites["site3"].res_idxs_pairs[0], [ 69, 852])

    def test_sites_consensus(self):
        res = cli.sites([test_filenames.tip_consensus_json],
                        [self.traj],
                        self.geom,
                        GPCR_UniProt=test_filenames.adrb2_human_xlsx,
                        CGN_UniProt=test_filenames.gnas2_human_xlsx,
                        no_disk=True, figures=False, accept_guess=True)
        res : mdciao.contacts.ContactGroup = res["interesting contacts"]
        self.assertSequenceEqual(res.ctc_labels_w_fragments_short_AA,
                                 ['L394@G.H5.26-K270@6.32x32',
                                  'D381@G.H5.13-Q229@5.68x68',
                                  'Q384@G.H5.16-Q229@5.68x68',
                                  'R385@G.H5.17-Q229@5.68x68',
                                  'D381@G.H5.13-K232@5.71x71',
                                  'Q384@G.H5.16-I135@3.54x54']
                                 )


class Test_interface(TestCLTBaseClass):

    @classmethod
    def setUpClass(cls):
        super(Test_interface, cls).setUpClass()
        cls.no_disk = True  # doesn't seem to speed things up much

    def test_interface(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.interface([self.traj, self.traj_reverse],
                          self.geom,
                          interface_selection_1=[0],
                          interface_selection_2=[1],
                          output_dir=tmpdir,
                          flareplot=True,
                          no_disk=self.no_disk
                          )

    def test_no_top(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.interface([self.traj, self.traj_reverse],
                          interface_selection_1=[0],
                          interface_selection_2=[1],
                          output_dir=tmpdir,
                          flareplot=False,
                          plot_timedep=False,
                          no_disk = self.no_disk
                          )

    def test_interface_wo_frag_idxs_groups(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            input_values = (val for val in ["0-1", "2,3"])
            with mock.patch('builtins.input', lambda *x: next(input_values)):
                cli.interface([self.traj, self.traj_reverse],
                              self.geom,
                              output_dir=tmpdir,
                              flareplot=False,
                              no_disk=self.no_disk
                              )

    def test_w_just_one_fragment_by_user(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.interface([self.traj, self.traj_reverse],
                          self.geom,
                          output_dir=tmpdir,
                          fragments=["0-5"],
                          flareplot=False,
                          no_disk=self.no_disk
                          )

    def test_w_just_one_fragment_by_user_and_n_ctcs(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.interface([self.traj, self.traj_reverse],
                          self.geom,
                          output_dir=tmpdir,
                          fragments=["0-5"],
                          n_nearest=1,
                          flareplot=False,
                          no_disk=self.no_disk
                          )

    def test_w_just_two_fragments_by_user(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            cli.interface([self.traj, self.traj_reverse],
                          self.geom,
                          interface_selection_1=[0],
                          interface_selection_2=[1],
                          output_dir=tmpdir,
                          fragments=["0-5",
                                     "6-10"],
                          flareplot=False,
                          no_disk=self.no_disk
                          )

    def test_w_nomenclature_CGN_GPCR(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            shutil.copy(test_filenames.adrb2_human_xlsx, tmpdir)
            with remember_cwd():
                os.chdir(tmpdir)
                cli.interface([self.traj, self.traj_reverse],
                              self.geom,
                              output_dir=tmpdir,
                              fragments=["967-1001",  # TM6
                                         "328-353"],  # a5
                              CGN_UniProt="gnas2_human",
                              GPCR_UniProt="adrb2_human",
                              accept_guess=True,
                              flareplot=False,
                              no_disk=self.no_disk
                              )

    def test_w_nomenclature_CGN_GPCR_fragments_are_consensus(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            input_values = (val for val in ["TM6", "*H5"])
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            shutil.copy(test_filenames.adrb2_human_xlsx, tmpdir)
            with remember_cwd():
                os.chdir(tmpdir)
                with mock.patch('builtins.input', lambda *x: next(input_values)):
                    cli.interface([self.traj, self.traj_reverse],
                                  self.geom,
                                  output_dir=tmpdir,
                                  fragments=["consensus"],
                                  CGN_UniProt="gnas2_human",
                                  GPCR_UniProt="adrb2_human",
                                  accept_guess=True,
                                  flareplot=False,
                                  no_disk=self.no_disk
                                  )

    def test_w_nomenclature_CGN_GPCR_fragments_are_consensus_and_flareplot(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            input_values = (val for val in ["TM6", "*H5"])
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            shutil.copy(test_filenames.adrb2_human_xlsx, tmpdir)
            with remember_cwd():
                os.chdir(tmpdir)
                with mock.patch('builtins.input', lambda *x: next(input_values)):
                    cli.interface([self.traj, self.traj_reverse],
                                  self.geom,
                                  output_dir=tmpdir,
                                  fragments=["consensus"],
                                  CGN_UniProt="gnas2_human",
                                  GPCR_UniProt="adrb2_human",
                                  accept_guess=True,
                                  #no_disk=self.no_disk
                                  )


    def test_w_nomenclature_CGN_GPCR_fragments_are_consensus_and_flareplot_and_self_interface(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            shutil.copy(test_filenames.adrb2_human_xlsx, tmpdir)
            with remember_cwd():
                os.chdir(tmpdir)
                cli.interface([self.traj, self.traj_reverse],
                              self.geom,
                              ctc_cutoff_Ang=5,
                              n_nearest=4,
                              output_dir=tmpdir,
                              fragments=["consensus"],
                              CGN_UniProt="gnas2_human",
                              GPCR_UniProt="adrb2_human",
                              accept_guess=True,
                              interface_selection_1='TM6',
                              interface_selection_2='TM5,TM6',
                              self_interface=True,
                              )

    def test_w_nomenclature_CGN_GPCR_fragments_are_consensus_and_flareplot_and_AA_selection_OR(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            shutil.copy(test_filenames.adrb2_human_xlsx, tmpdir)
            with remember_cwd():
                os.chdir(tmpdir)
                intf = cli.interface([self.traj, self.traj_reverse],
                                     self.geom,
                                     ctc_cutoff_Ang=5,
                                     n_nearest=4,
                                     output_dir=tmpdir,
                                     fragments=["consensus"],
                                     CGN_UniProt="gnas2_human",
                                     GPCR_UniProt="adrb2_human",
                                     accept_guess=True,
                                     interface_selection_1='TM6',
                                     interface_selection_2='TM5',
                                     self_interface=True,
                                     AA_selection="5.50x50-5.55x55"
                                     )
            TM5 = ["5.50x50", "5.51x51", "5.52x52", "5.53x53", "5.54x54", "5.55x55"]
            assert all ([lab[1] in TM5 for lab in  intf.consensus_labels]), intf.consensus_labels
            _plt.close("all")

    def test_w_nomenclature_CGN_GPCR_fragments_are_consensus_and_flareplot_and_AA_selection_AND(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            shutil.copy(test_filenames.adrb2_human_xlsx, tmpdir)
            with remember_cwd():
                intf = cli.interface([self.traj, self.traj_reverse],
                                     self.geom,
                                     ctc_cutoff_Ang=5,
                                     n_nearest=4,
                                     output_dir=tmpdir,
                                     fragments=["consensus"],
                                     CGN_UniProt="gnas2_human",
                                     GPCR_UniProt="adrb2_human",
                                     accept_guess=True,
                                     interface_selection_1='TM6',
                                     interface_selection_2='TM5',
                                     self_interface=True,
                                     AA_selection=["5.50x50-5.55x55", "6.45x45,6.49x49"]
                                     )
            TM5 = ["5.50x50", "5.51x51", "5.52x52", "5.53x53", "5.54x54", "5.55x55"]
            assert all ([lab[1] in TM5 for lab in  intf.consensus_labels]), intf.consensus_labels
            assert all ([lab[0] in ["6.45x45","6.49x49"] for lab in  intf.consensus_labels]), intf.consensus_labels
            _plt.close("all")

    def test_w_nomenclature_CGN_GPCR_fragments_are_consensus_and_flareplot_and_AA_selection_raises(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            shutil.copy(test_filenames.adrb2_human_xlsx, tmpdir)
            with remember_cwd():
                with self.assertRaises(ValueError):
                    intf = cli.interface([self.traj, self.traj_reverse],
                                         self.geom,
                                         ctc_cutoff_Ang=5,
                                         n_nearest=4,
                                         output_dir=tmpdir,
                                         fragments=["consensus"],
                                         CGN_UniProt="gnas2_human",
                                         GPCR_UniProt="adrb2_human",
                                         accept_guess=True,
                                         interface_selection_1='TM6',
                                         interface_selection_2='TM5',
                                         self_interface=True,
                                         AA_selection=["5.50x50-5.55x55,6.45x45,6.49x49"]
                                         )




class Test_pdb(TestCLTBaseClass):

    def test_works(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            geom = cli.pdb("3SN6", filename=None)
            isinstance(geom,md.Trajectory)

class Test_parse_consensus_option(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.geom = md.load(test_filenames.top_pdb)

    def test_empty(self):
        residx2conlab= cli._parse_consensus_option(None, None, self.geom.top,
                                                   None,
                                                   )
        assert _pandasunique(_np.array(residx2conlab))[0] is None

    def test_empty_w_return(self):
        residx2conlab, lblr  = cli._parse_consensus_option(None, None, self.geom.top,
                                                           None,
                                                           return_Labeler=True)
        assert lblr is None
        assert _pandasunique(_np.array(residx2conlab))[0] is None

    def test_with_GPCR(self):
        fragments = mdcfragments.get_fragments(self.geom.top)
        input_values = (val for val in [""])
        option = test_filenames.adrb2_human_xlsx
        with mock.patch('builtins.input', lambda *x: next(input_values)):
            residx2conlab, lblr = cli._parse_consensus_option(option, "GPCR",
                                                              self.geom.top,
                                                              fragments,
                                                              return_Labeler=True,
                                                              try_web_lookup=False)
            self.assertIsInstance(lblr, LabelerGPCR)
            self.assertIsInstance(residx2conlab,list)

    def test_with_GPCR_already_instantiated(self):
        fragments = mdcfragments.get_fragments(self.geom.top)
        GPCR = LabelerGPCR(test_filenames.adrb2_human_xlsx)
        input_values = (val for val in [""])
        with mock.patch('builtins.input', lambda *x: next(input_values)):

            residx2conlab, lblr = cli._parse_consensus_option(GPCR, "GPCR",
                                                              self.geom.top,
                                                              fragments,
                                                              return_Labeler=True)
            self.assertIsInstance(lblr, LabelerGPCR)
            self.assertEqual(lblr,GPCR)
            self.assertIsInstance(residx2conlab,list)

    def test_with_GPCR_labels(self):
        fragments = mdcfragments.get_fragments(self.geom.top)
        GPCR = LabelerGPCR(test_filenames.adrb2_human_xlsx)
        GPCR_labels = GPCR.top2labels(self.geom.top)
        residx2conlab, lblr = cli._parse_consensus_option(GPCR_labels, "GPCR",
                                                              self.geom.top,
                                                              fragments,
                                                              return_Labeler=True)
        assert lblr is None
        assert residx2conlab is GPCR_labels
    def test_with_GPCR_labels_raises(self):
        fragments = mdcfragments.get_fragments(self.geom.top)
        GPCR = LabelerGPCR(test_filenames.adrb2_human_xlsx)
        GPCR_labels = GPCR.top2labels(self.geom.top)
        with self.assertRaises(AssertionError):
            residx2conlab, lblr = cli._parse_consensus_option(GPCR_labels[:20], "GPCR",
                                                              self.geom.top,
                                                              fragments,
                                                              return_Labeler=True)

    def test_no_answer(self):
        GPCR = LabelerGPCR(test_filenames.adrb2_human_xlsx)
        residx2conlab, lblr = cli._parse_consensus_option(GPCR, "GPCR",
                                                          self.geom.top,
                                                          [_np.arange(10)],
                                                          return_Labeler=True)

class Test_offer_to_create_dir(unittest.TestCase):

    def test_creates_dir(self):
       with TemporaryDirectory() as tmpdir:
           newdir = _path.join(tmpdir,"non_existent")
           input_values = (val for val in [""])
           with mock.patch('builtins.input', lambda *x: next(input_values)):
               cli._offer_to_create_dir(newdir)
           assert _path.exists(newdir)

    def test_does_nothing_bc_dir_exists(self):
       with TemporaryDirectory() as tmpdir:
           cli._offer_to_create_dir(tmpdir)

    def test_raises(self):
        input_values = (val for val in ["n"])
        with mock.patch('builtins.input', lambda *x: next(input_values)):
            res = cli._offer_to_create_dir("testdir")
            assert res is None

class Test_load_any_geom(unittest.TestCase):

    def test_loads_file(self):
        geom = cli._load_any_geom(test_filenames.top_pdb)
        self.assertIsInstance(geom, md.Trajectory)

    def test_loads_geom(self):
        geom = md.load(test_filenames.top_pdb)
        geomout = cli._load_any_geom(geom)
        self.assertIs(geom, geomout)

class Test_parse_fragment_naming_options(unittest.TestCase):
    def setUp(self):
        self.fragments = [[0,1],
                          [2,3],
                          [4,5],
                          [6,7]]
    def test_empty_str(self):
        fragnames = cli._parse_fragment_naming_options("", self.fragments)
        self.assertSequenceEqual(["frag0","frag1","frag2","frag3"],
                                 fragnames)

    def test_None(self):
        fragnames = cli._parse_fragment_naming_options("None", self.fragments)
        self.assertSequenceEqual([None,None,None,None],
                                 fragnames)

    def test_list(self):
        fragnames = cli._parse_fragment_naming_options(["A","B","C","D"], self.fragments)
        self.assertSequenceEqual(["A","B","C","D"],
                                 fragnames)
    def test_csv(self):
        fragnames = cli._parse_fragment_naming_options("TM1,TM2,ICL3,H8", self.fragments)
        self.assertSequenceEqual(["TM1","TM2","ICL3","H8"],
                                 fragnames)

    def test_csv_wrong_nr_raises(self):
        with self.assertRaises(AssertionError):
            fragnames = cli._parse_fragment_naming_options("TM1,TM2", self.fragments)
            self.assertSequenceEqual(["TM1", "TM2", "ICL3", "H8"],
                                     fragnames)
    def test_danger_raises(self):
        with self.assertRaises(NotImplementedError):
            cli._parse_fragment_naming_options("TM1,danger", self.fragments)

class Test_parse_parse_coloring_options(unittest.TestCase):

    def test_none(self):
        color = cli._parse_coloring_options(None,2)
        self.assertSequenceEqual(["tab:blue","tab:blue"], color)

    def test_True(self):
        color = cli._parse_coloring_options(True,2)
        self.assertSequenceEqual(['magenta', 'yellow'], color)

    def test_string(self):
        color = cli._parse_coloring_options("salmon",2)
        self.assertSequenceEqual(['salmon', 'salmon'], color)

    def test_list(self):
        color = cli._parse_coloring_options(["pink", "salmon"],2)
        self.assertSequenceEqual(["pink", "salmon"], color)

    def test_raises(self):
        with self.assertRaises(ValueError):
            color = cli._parse_coloring_options(["pink", "salmon"],3)

class Test_color_schemes(unittest.TestCase):

    def test_scheme_P(self):
        self.assertSequenceEqual(cli._color_schemes("P"),["red", "purple", "gold", "darkorange"])

    def test_auto(self):
        self.assertEqual(len(cli._color_schemes("auto")),10)

    def test_color(self):
        self.assertEqual(cli._color_schemes("blue"),["blue"])

    def test_csv(self):
        self.assertEqual(cli._color_schemes("blue,red"),["blue","red"])


class Test_fragment_overview(unittest.TestCase):

    def test_runs(self):
        res = cli.fragment_overview(test_filenames.top_pdb)
        self.assertSequenceEqual(list(res.keys()),
                                 mdcfragments.fragments._allowed_fragment_methods)

class Test_fragment_overview_Nomenclature(unittest.TestCase):

    def test_CGN_paths(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            path_to_gnas2_human = _path.join(tmpdir, _path.basename(test_filenames.gnas2_human_xlsx))
            shutil.copy(test_filenames.gnas2_human_xlsx, tmpdir)
            with remember_cwd():
                os.chdir(tmpdir)
                a = parser_for_CGN_overview()
                a = a.parse_args([path_to_gnas2_human])
                a.__setattr__("topology",test_filenames.top_pdb)
                cli._fragment_overview(a, "CGN")

    def test_GPCR_paths_and_verbose(self):
        a = parser_for_GPCR_overview()
        a = a.parse_args([test_filenames.adrb2_human_xlsx])
        a.__setattr__("print_conlab",True)
        a.__setattr__("topology",test_filenames.top_pdb)
        cli._fragment_overview(a, "GPCR")

    def test_GPCR_url(self):
        a = parser_for_GPCR_overview()
        a = a.parse_args(["adrb2_human"])
        a.__setattr__("topology",test_filenames.pdb_3SN6)
        cli._fragment_overview(a, "GPCR")

    def test_raises(self):
        with self.assertRaises(ValueError):
            cli._fragment_overview(None, "BWx")

    def test_AAs(self):
        a = parser_for_CGN_overview()
        a = a.parse_args([test_filenames.gnas2_human_xlsx,
                          ])
        a.__setattr__("AAs","LEU394,LEU395")
        a.__setattr__("topology",test_filenames.top_pdb)
        cli._fragment_overview(a, "CGN")

    def test_labels(self):
        a = parser_for_GPCR_overview()
        a = a.parse_args([test_filenames.adrb2_human_xlsx])
        a.__setattr__("topology",test_filenames.top_pdb)
        a.__setattr__("labels","3.50")
        cli._fragment_overview(a, "GPCR")

    def test_no_top(self):
        a = parser_for_GPCR_overview()
        a = a.parse_args([test_filenames.adrb2_human_xlsx])
        a.__setattr__("labels","3.50")
        cli._fragment_overview(a, "GPCR")

class Test_compare(unittest.TestCase):

    def setUp(self):
        self.CG1 = contacts.ContactGroup([contacts.ContactPair([0,1],[[.1, .1]], [[0., 1.]]),
                                 contacts.ContactPair([0,2],[[.1, .2]], [[0., 1.]])])

        self.CG2 = contacts.ContactGroup([contacts.ContactPair([0,1],[[.1, .1, .1]], [[0., 1., 2.]]),
                                 contacts.ContactPair([0,3],[[.1, .2, .2]], [[0., 1., 2.]])])

    def test_just_works(self):
        with TemporaryDirectory(suffix='_test_mdciao') as tmpdir:
            with remember_cwd():
                os.chdir(tmpdir)
                myfig, freqs, plotted_freqs = cli.compare({"CG1": self.CG1, "CG2": self.CG2},
                                       ctc_cutoff_Ang=1.5, anchor="0")
                myfig.tight_layout()
                assert isinstance(freqs, dict)
                assert isinstance(plotted_freqs, dict)
                _plt.close("all")

if __name__ == '__main__':
    unittest.main()

class Test_residue_selection(unittest.TestCase):

    def test_works(self):
        residue_idxs, __, maps = cli.residue_selection("GLU30",test_filenames.small_monomer)

        _np.testing.assert_array_equal(residue_idxs,[0])

        self.assertDictEqual(maps, {})



import tempfile

import numpy as np

from mdciao.utils.COM import *
from mdciao.utils.COM import _unwrap, _per_residue_unwrapping
import mdtraj as md
import numpy as _np
import unittest
from mdciao.examples import filenames as test_filenames

class Test_COM_utils(unittest.TestCase):

    def setUp(self):
        super(Test_COM_utils,self).setUp()
        self.pdb_file = test_filenames.top_pdb
        self.file_xtc = test_filenames.traj_xtc_stride_20
        self.top = md.load(self.pdb_file).top
        self.traj = md.load(self.file_xtc, top=self.top)

        self.traj_5_frames = self.traj[:5]
        # Very slow, but what other way of directly computing the COM is there?
        COMS_mdtraj = [md.compute_center_of_mass(self.traj_5_frames.atom_slice([aa.index for aa in rr.atoms])) for rr in
                       self.top.residues]
        # re order along the time axis
        self.COMS_mdtraj = _np.zeros((5, self.top.n_residues, 3))
        for ii in range(5):
            self.COMS_mdtraj[ii,:, :] = [jcom[ii] for jcom in COMS_mdtraj]

    def test_COMxyz_works(self):
        COMSs_mine = geom2COMxyz(self.traj_5_frames)
        _np.testing.assert_allclose(self.COMS_mdtraj, COMSs_mine)

    def test_COMxyz_works_some_residues(self):
        residue_idxs = [1, 3, 5, 7]
        COMSs_mine = geom2COMxyz(self.traj_5_frames, residue_idxs=residue_idxs)

        _np.testing.assert_allclose(COMSs_mine,
                                    self.COMS_mdtraj[:,residue_idxs])

    def test_COMdist_works(self):
        res_pairs = [[0,10], [10,20]]
        Dref = _np.vstack((_np.linalg.norm(self.COMS_mdtraj[:,0]-self.COMS_mdtraj[:,10], axis=1),
                           _np.linalg.norm(self.COMS_mdtraj[:,10]- self.COMS_mdtraj[:,20], axis=1))).T
        COMdist =  geom2COMdist(self.traj_5_frames, residue_pairs=res_pairs)
        _np.testing.assert_array_almost_equal(Dref, COMdist)

class Test_geom2max_residue_radius(unittest.TestCase):

    def setUp(self) -> None:
        pdb = "CRYST1   89.862   89.862  142.612  90.00  90.00  90.00 P 1           1 \n" \
              "MODEL        0 \n" \
              "ATOM      1  CA  GLU A  30      0.0000  0.0000  0.0000  1.00  0.00           C \n" \
              "ATOM      2  CB  GLU A  30      2.0000  0.0000  0.0000  1.00  0.00           C \n" \
              "ATOM      3  C   GLU A  30      10.000  0.0000  0.0000  1.00  0.00           C \n" \
              "ATOM      4  CA  VAL A  31      0.0000  0.0000  0.0000  1.00  0.00           C \n" \
              "ATOM      5  CB  VAL A  31      0.0000  5.0000  0.0000  1.00  0.00           C \n" \
              "ATOM      6  C   VAL A  31      0.0000  10.000  0.0000  1.00  0.00           C \n" \
              "ATOM      7  CA  TRP A  32      0.0000  0.0000  0.0000  1.00  0.00           C \n" \
              "ATOM      8  CB  TRP A  32      0.0000  0.0000  8.0000  1.00  0.00           C \n" \
              "ATOM      9  C   TRP A  32      0.0000  0.0000  10.000  1.00  0.00           C \n" \
              "TER      10      TRP A  32 "
        with tempfile.NamedTemporaryFile(suffix=".pdb") as tf:
            with open(tf.name, "w") as f:
                f.write(pdb)
            self.geom = md.load(tf.name)
            self.geom = self.geom.join(self.geom)
            self.geom._xyz[1,:,:] *= 10
        # With the above geometry, it's easy to see that the COMs are, for x,y,z respectively
        # GLU30 (0+2+10)/3 = 4
        # VAL31 (0+5+10)/3 = 5
        # TRP31 (0+8+10)/3 = 6
        # And then the radii are, respectively
        # GLU30 10 - 4 = 6
        # VAL31 10 - 5 = 5 or 0 - 5 = 5
        # TRP31  0 - 6 = -6 in abs 6
        # Note that we have to multiply x .1 to be back in nm (pdb strg is in Ang)
        # and that the second frame is the first multiplied by 10

    def test_works(self):
        rrm = geom2max_residue_radius(self.geom)
        _np.testing.assert_array_almost_equal(rrm, _np.array([[.6, .5, .6],
                                                              [6, 5, 6]]))

    def test_w_residxs(self):
        rrm = geom2max_residue_radius(self.geom, residue_idxs=[1, 2])
        _np.testing.assert_array_almost_equal(rrm, _np.array([[.5, .6],
                                                              [5, 6]]))

    def test_w_residxs_and_resCOM(self):
        rrm = geom2max_residue_radius(self.geom, residue_idxs=[1, 2], res_COMs=_np.array([[[0, .5, 0],
                                                                                           [0, 0, .6]],
                                                                                          [[0, 5, 0],
                                                                                           [0, 0, 6]]
                                                                                          ]))
        _np.testing.assert_array_almost_equal(rrm, _np.array([[.5, .6],
                                                              [5, 6]]))

    def test_raises(self):
        with self.assertRaises(ValueError):
            geom2max_residue_radius(self.geom, res_COMs=_np.array([[[0, .5, 0],
                                                                    [0, 0, .6]],
                                                                   [[0, 5, 0],
                                                                    [0, 0, 6]]
                                                                   ]))

    def test_COMdist_subtract_radii_low_and_high_mem(self):
        res_coms = _np.array([[4,0,0],
                              [0,5,0],
                              [0,0,6]], dtype=float) #<- we know this from the definition of the geom
        res_coms *= .1 #convert to nm
        res_coms = _np.array([res_coms, res_coms*10]) # add a second frame
        _np.testing.assert_array_almost_equal(res_coms, geom2COMxyz(self.geom, [0,1,2])) # testing geom2COM on-the-fly

        d_0_1 = _np.sqrt(4**2+5**2)*.1
        d_0_2 = _np.sqrt(4**2+6**2)*.1
        d_1_2 = _np.sqrt(5**2+6**2)*.1
        d_0_1 = _np.array((d_0_1, d_0_1 * 10))
        d_0_2 = _np.array((d_0_2, d_0_2 * 10))
        d_1_2 = _np.array((d_1_2, d_1_2 * 10))
        comD = geom2COMdist(self.geom, [[0, 1], [0, 2], [1, 2]], periodic=False, per_residue_unwrap=False )
        _np.testing.assert_array_almost_equal(comD, _np.array([d_0_1, d_0_2, d_1_2]).T) # testing geom2COMdist w/o subtract_radii

        # Test low-mem
        # We subtract the same max-radii by hand (which we know as rrm from test_works)
        # for all frames of comD
        comD[:, 0] -= (6 + 5)
        comD[:, 1] -= (6 + 6)
        comD[:, 2] -= (5 + 6)

        # This tests low_mem
        comDr = geom2COMdist(self.geom, [[0, 1], [0, 2], [1, 2]], subtract_max_radii=True, periodic=False, per_residue_unwrap=False)
        _np.testing.assert_array_almost_equal(comDr, comD)  # the actual test

        # Test high_mem
        comD = geom2COMdist(self.geom, [[0, 1], [0, 2], [1, 2]],  periodic=False, per_residue_unwrap=False)
        # We subtract the per-frame max-radii by hand (which we know as rrm from test_works)
        comD[0][0] -= (.6 + .5)
        comD[0][1] -= (.6 + .6)
        comD[0][2] -= (.5 + .6)
        comD[1][0] -= (6 + 5)
        comD[1][1] -= (6 + 6)
        comD[1][2] -= (5 + 6)

        # This test high_mem
        comDr = geom2COMdist(self.geom, [[0, 1], [0, 2], [1, 2]], subtract_max_radii=True, low_mem=False, periodic=False, per_residue_unwrap=False)
        _np.testing.assert_array_almost_equal(comDr, comD)  # the actual test

class TestUnwrapping(unittest.TestCase):

    def setUp(self) -> None:
        pdb = "CRYST1   100.00   100.00  100.000  90.00  90.00  90.00 P 1           1 \n" \
              "MODEL        0 \n" \
              "ATOM      1  CA  GLU A  30      5.0000  30.000  10.000  1.00  0.00           C \n" \
              "ATOM      2  CB  GLU A  30      90.000  95.000  10.000  1.00  0.00           C \n" \
              "ATOM      3  CA  VAL A  31      0.0000  0.0000  0.0000  1.00  0.00           C \n" \
              "ATOM      4  CB  VAL A  31      0.0000  5.0000  0.0000  1.00  0.00           C \n" \
              "TER       4      VAL A  31 "
        with tempfile.NamedTemporaryFile(suffix=".pdb") as tf:
            with open(tf.name, "w") as f:
                f.write(pdb)
            self.geom = md.load(tf.name)

    def test_unwrap_coords(self):
        unitcell_lengths = _np.array([[10, 10, 10]])


        wrapped_coords = _np.array([[[.5, 3, 1], # <- This atom is closest to the center of the box, its coords won't change
                                     [9, 9.5, 1]]])

        un_wrapped_coords_ref = _np.array([[[.5, 3, 1],
                                            [-1, -0.5, 1]]])

        _unwrapped_coords = _unwrap(wrapped_coords, unitcell_lengths)
        _np.testing.assert_array_equal(_unwrapped_coords,
                                       un_wrapped_coords_ref
                                       )

    def test_unwrap_residues(self):
        per_res_unwrapped = _per_residue_unwrapping(self.geom)

        _np.testing.assert_array_equal(per_res_unwrapped.xyz, _np.array([[[.5, 3, 1],
                                                                          [-1, -0.5, 1],
                                                                          [0, 0, 0],
                                                                          [0, .5, 0]]]))

        assert per_res_unwrapped is not self.geom

    def test_unwrap_residues_residue_idxs(self):
        # will leave geom untouched bc residue 1 does not need unwrapping
        per_res_unwrapped = _per_residue_unwrapping(self.geom, residue_idxs=[1])

        _np.testing.assert_array_equal(per_res_unwrapped.xyz, _np.array([[[.5, 3, 1],
                                                                          [9, 9.5, 1],
                                                                          [0, 0, 0],
                                                                          [0, .5, 0]]]))

        # will have only unwrapped residue 0
        per_res_unwrapped = _per_residue_unwrapping(self.geom, residue_idxs=[0])

        _np.testing.assert_array_equal(per_res_unwrapped.xyz, _np.array([[[.5, 3, 1],
                                                                          [-1, -0.5, 1],
                                                                          [0, 0, 0],
                                                                          [0, .5, 0]]]))

        assert per_res_unwrapped is not self.geom

    def test_inplace_w_maxradii(self):
        geom = md.Trajectory(_np.copy(self.geom._xyz), self.geom.top, time=self.geom.time, unitcell_angles=self.geom.unitcell_angles,
                       unitcell_lengths=self.geom.unitcell_lengths)
        per_res_unwrapped = _per_residue_unwrapping(geom, inplace=True,
                                                    max_res_radii_t=geom2max_residue_radius(geom))
        _np.testing.assert_array_equal(per_res_unwrapped.xyz, _np.array([[[.5, 3, 1],
                                                                          [-1, -0.5, 1],
                                                                          [0, 0, 0],
                                                                          [0, .5, 0]]]))

        assert per_res_unwrapped is geom

    def test_COM_dist_wo_unwrapping_wo_subtract_wo_periodic(self):
        COM1 = np.array([5 + 90, 30 + 95, 10 + 10]) / 2 / 10    # [4.75 6.25 1.  ]
        COM2 = np.array([0 + 0, 0 + 5, 0 + 0]) / 2 / 10         # [0.   0.25 0.  ]
        D12 = np.linalg.norm(COM1 - COM2)

        np.testing.assert_array_equal(D12,
                                      geom2COMdist(self.geom, [[0, 1]],
                                                   subtract_max_radii=False, low_mem=True, periodic=False,
                                                   per_residue_unwrap=False))

    def test_COM_dist_wo_unwrapping_wo_subtract_w_periodic(self):
        # PBCs
        dx = 4.75
        dy = 6  # This is larger than half the box, hence pick the periodic image
        dy -= self.geom.unitcell_lengths[0, 1]
        dz = 1
        np.testing.assert_array_equal(np.sqrt(dx ** 2 + dy ** 2 + dz ** 2),
                                      geom2COMdist(self.geom, [[0, 1]],
                                                   subtract_max_radii=False, low_mem=True, periodic=True,
                                                   per_residue_unwrap=False))

    def test_COM_dist_w_unwrapping_wo_subtract(self):
        """
        # These we take from above
        unwrapped_coords = _np.array([[[.5, 3, 1],
                                     [-1, -0.5, 1],
                                     [0, 0, 0],
                                     [0, .5, 0]]])

        COM1 = np.array([.5 - 1, 3 - .5, 1 + 1]) / 2    # [-0.25  1.25  1.  ]
        COM2 = np.array([0 + 0, 0 + .5, 0 + 0]) / 2     # [0.  .25 0. ]
        """
        dx = .25
        dy = 1
        dz = 1

        # None of the deltas is larger than half the box, no correction needed
        np.testing.assert_almost_equal(np.sqrt(dx ** 2 + dy ** 2 + dz ** 2),
                                       geom2COMdist(self.geom, [[0, 1]],
                                                    subtract_max_radii=False, low_mem=True, periodic=True,
                                                    per_residue_unwrap=True))

    def test_COM_dist_w_unwrapping_w_subtract_low_and_hi_mem(self):
        # The COMs we take from above.
        COM1, COM2 = _np.array([-0.25, 1.25, 1.]), _np.array([0., .25, 0.])
        # Since it's only 2 atoms per residue, any atom of the residue can be used for the max residue radius
        max_r_1 = _np.linalg.norm(COM1 - np.array([.5, 3, 1]))
        max_r_2 = _np.linalg.norm(COM2 - np.array([0, 0, 0]))
        # None of the deltas is larger than half the box, no correction needed
        D12 = np.linalg.norm(COM1 - COM2) - max_r_1 - max_r_2
        np.testing.assert_almost_equal(D12,
                                       geom2COMdist(self.geom, [[0, 1]],
                                                    subtract_max_radii=True, low_mem=True, periodic=True,
                                                    per_residue_unwrap=True))

        np.testing.assert_almost_equal(D12,
                                       geom2COMdist(self.geom, [[0, 1]],
                                                    subtract_max_radii=True, low_mem=False, periodic=True,
                                                    per_residue_unwrap=True))

if __name__ == '__main__':
    unittest.main()


import mdtraj as md
from os import path
import numpy as np

import unittest
from mdciao.examples import filenames as test_filenames

from mdciao.utils import str_and_dict

from tempfile import TemporaryDirectory as _TDir

import io as _io, contextlib as _contextlib

from pandas import \
    DataFrame as _DF,\
    ExcelWriter as _XW

class Test_get_sorted_trajectories(unittest.TestCase):
    def setUp(self):
        self.geom = md.load(test_filenames.top_pdb)
        self.traj = md.load(test_filenames.traj_xtc_stride_20, top=self.geom.top)
        self.traj_reverse = md.load(test_filenames.traj_xtc_stride_20, top=self.geom.top)[::-1]

    def test_glob_with_pattern(self):
        str_and_dict.get_trajectories_from_input(path.join(test_filenames.example_path, "*.xtc"))

    def test_glob_with_filename(self):
        str_and_dict.get_trajectories_from_input(test_filenames.traj_xtc_stride_20)

    def test_with_one_trajectory_object(self):
        list_out = str_and_dict.get_trajectories_from_input(self.traj)
        assert len(list_out)==1
        assert isinstance(list_out[0], md.Trajectory)

    def test_with_trajectory_objects(self):
        str_and_dict.get_trajectories_from_input([self.traj,
                                                  self.traj_reverse])


    def test_fails_if_not_traj_at_all(self):
        with self.assertRaises(FileNotFoundError):
            str_and_dict.get_trajectories_from_input("bogus.xtc")

    def test_fails_if_not_trajs(self):
        with self.assertRaises(AssertionError):
            str_and_dict.get_trajectories_from_input([self.traj,
                                                      1])


class Test_inform_about_trajectories(unittest.TestCase):
    def setUp(self):
        self.geom = md.load(test_filenames.top_pdb)
        self.traj = md.load(test_filenames.traj_xtc_stride_20, top=self.geom.top)
        self.traj_reverse = md.load(test_filenames.traj_xtc_stride_20, top=self.geom.top)[::-1]

    def test_fails_no_list(self):
        with self.assertRaises(AssertionError):
            str_and_dict.inform_about_trajectories(test_filenames.traj_xtc_stride_20)

    def test_list_of_files(self):
        str_and_dict.inform_about_trajectories([test_filenames.traj_xtc_stride_20,
                                                test_filenames.traj_xtc_stride_20])

    def test_list_of_trajs(self):
        str_and_dict.inform_about_trajectories([self.traj,
                                                self.traj_reverse])


class Test_replace_w_dict(unittest.TestCase):
    outkey = str_and_dict.replace_w_dict("key I don't like", {"key": "word", "like": "love"})
    assert outkey=="word I don't love"

class Test_delete_exp_inkeys(unittest.TestCase):

    def test_runs(self):
        indict = {"GLU30-LYS40":True,
                  "ARG40-GLU30":False}

        outdict, deleted = str_and_dict.delete_exp_in_keys(indict, "GLU30")
        self.assertEqual(deleted,["GLU30","GLU30"])
        assert len(outdict)==2
        assert outdict["LYS40"]
        assert not outdict["ARG40"]


class Test_iterate_and_inform_lambdas(unittest.TestCase):

    def setUp(self):
        self.filename = test_filenames.traj_xtc_stride_20
        self.pdb = test_filenames.top_pdb
        self.top = md.load(self.pdb).top
        self.traj = md.load(self.filename, top=self.top)
        self.stride=3

    def _call_iterators_and_test_them(self, iterate, inform, ixtc, stride=1):
        print()
        times = []
        nf = 0
        for ii, chunk in enumerate(iterate(ixtc)):
            times.append(chunk.time)
            nf += chunk.n_frames
            inform(ixtc, 0, ii, nf)
            print()
        times = np.hstack(times)
        # This is indirectly checking the right frames were grabbed
        assert np.allclose(times, self.traj.time[::stride])
        print()

    def test_filename_no_stride(self):
        iterate, inform = str_and_dict.iterate_and_inform_lambdas(self.filename,
                                                     10,
                                                     top=self.top)
        self._call_iterators_and_test_them(iterate, inform, self.filename, 1)

    def test_filename_no_stride_filename_is_ascii_just_works(self):
        iterate, inform = str_and_dict.iterate_and_inform_lambdas(self.pdb,
                                                     10,
                                                     top=self.top)
        nf = 0
        for ii, chunk in enumerate(iterate(self.pdb)):
            nf += chunk.n_frames
            inform(self.pdb,0,ii,nf)
        assert nf==1
        assert ii==0


    def test_filename_w_stride(self):
        iterate, inform = str_and_dict.iterate_and_inform_lambdas(self.filename,
                                                     10,
                                                     stride=self.stride,
                                                     top=self.top)

        self._call_iterators_and_test_them(iterate, inform, self.filename,
                                           stride = self.stride)

    def test_traj_no_stride(self):
        iterate, inform = str_and_dict.iterate_and_inform_lambdas(self.traj,
                                                     10,
                                                     )

        self._call_iterators_and_test_them(iterate, inform, self.traj)

    def test_traj_w_stride(self):
        iterate, inform = str_and_dict.iterate_and_inform_lambdas(self.traj,
                                                     10,
                                                     stride=self.stride
                                                     )

        self._call_iterators_and_test_them(iterate, inform, self.traj,
                                           stride=self.stride)

class Test_unify_freq_dicts(unittest.TestCase):

    def setUp(self):
        self.freq_dicts = {"WT":   {"GLU30-LYS40":1.0,
                                    "GLU30-LYS50":.5,
                                    "LYS60-GLU30": .1},
                           "K40A": {"ALA40-GLU30":.7,
                                   #"LYS50-GLU30":.5,
                                    "GLU30-LYS60": .2},
                           "prot": {"LYS40-GLH30": .3
                                    }}
        self.freq_dicts_frags = {"WT": {"GLU30@frag1-LYS40@frag2": 1.0,
                                        "GLU30@frag1-LYS50@frag3": .5,
                                        "LYS60@frag4-GLU30@frag1": .1},
                                 "K40A": {"ALA40-GLU30@frag1": .7,
                                          # "LYS50-GLU30":.5,
                                          "GLU30@frag1-LYS60@frag4": .2},
                                 "prot": {"LYS40@frag2-GLH30@frag1": .3
                                          }}
    def test_basic(self):
        out_dict = str_and_dict.unify_freq_dicts(self.freq_dicts,
                                    key_separator=None)
        #original
        assert out_dict["WT"]["GLU30-LYS40"] == 1.0
        assert out_dict["WT"]["GLU30-LYS50"] == .5
        assert out_dict["WT"]["LYS60-GLU30"] == .1
        #external
        assert out_dict["WT"]["ALA40-GLU30"] == 0 #from K40A
        assert out_dict["WT"]["GLU30-LYS60"] == 0 #from K40A
        assert out_dict["WT"]["LYS40-GLH30"] == 0 #from prot

        # original
        assert out_dict["K40A"]["ALA40-GLU30"] == .7
        assert out_dict["K40A"]["GLU30-LYS60"] == .2
        # external
        assert out_dict["K40A"]["GLU30-LYS40"] == 0 #from WT
        assert out_dict["K40A"]["GLU30-LYS50"] == 0 #from WT
        assert out_dict["K40A"]["LYS40-GLH30"] == 0 #from prot

        # original
        assert out_dict["prot"]["LYS40-GLH30"] == .3
        # external
        assert out_dict["prot"]["GLU30-LYS40"] == 0  # from WT
        assert out_dict["prot"]["GLU30-LYS50"] == 0  # from WT
        assert out_dict["prot"]["LYS60-GLU30"] == 0  # from WT

        assert out_dict["prot"]["LYS60-GLU30"] == 0  # from K40A
        assert out_dict["prot"]["ALA40-GLU30"] == 0  # from K40A

    def test_basic_defrag(self):
        out_dict = str_and_dict.unify_freq_dicts(self.freq_dicts,
                                                 key_separator=None,
                                                 defrag="@")
        #original
        assert out_dict["WT"]["GLU30-LYS40"] == 1.0
        assert out_dict["WT"]["GLU30-LYS50"] == .5
        assert out_dict["WT"]["LYS60-GLU30"] == .1
        #external
        assert out_dict["WT"]["ALA40-GLU30"] == 0 #from K40A
        assert out_dict["WT"]["GLU30-LYS60"] == 0 #from K40A
        assert out_dict["WT"]["LYS40-GLH30"] == 0 #from prot

        # original
        assert out_dict["K40A"]["ALA40-GLU30"] == .7
        assert out_dict["K40A"]["GLU30-LYS60"] == .2
        # external
        assert out_dict["K40A"]["GLU30-LYS40"] == 0 #from WT
        assert out_dict["K40A"]["GLU30-LYS50"] == 0 #from WT
        assert out_dict["K40A"]["LYS40-GLH30"] == 0 #from prot

        # original
        assert out_dict["prot"]["LYS40-GLH30"] == .3
        # external
        assert out_dict["prot"]["GLU30-LYS40"] == 0  # from WT
        assert out_dict["prot"]["GLU30-LYS50"] == 0  # from WT
        assert out_dict["prot"]["LYS60-GLU30"] == 0  # from WT

        assert out_dict["prot"]["LYS60-GLU30"] == 0  # from K40A
        assert out_dict["prot"]["ALA40-GLU30"] == 0  # from K40A

    def test_basic_exclude(self):
        with self.assertRaises(NotImplementedError):
            out_dict = str_and_dict.unify_freq_dicts(self.freq_dicts,
                                        exclude=["LYS6"], #to check that is just a pattern matching, not full str match
                                    key_separator=None)

        """
        #original
        assert out_dict["WT"]["GLU30-LYS40"] == 1.0
        assert out_dict["WT"]["GLU30-LYS50"] == .5
        assert "LYS60-GLU30" not in out_dict["WT"].keys(), out_dict["WT"].keys()

        #external
        assert out_dict["WT"]["ALA40-GLU30"] == 0 #from K40A
        assert out_dict["WT"]["GLU30-LYS60"] == 0 #from K40A
        assert out_dict["WT"]["LYS40-GLH30"] == 0 #from prot

        # original
        assert out_dict["K40A"]["ALA40-GLU30"] == .7
        assert out_dict["K40A"]["GLU30-LYS60"] == .2
        # external
        assert out_dict["K40A"]["GLU30-LYS40"] == 0 #from WT
        assert out_dict["K40A"]["GLU30-LYS50"] == 0 #from WT
        assert out_dict["K40A"]["LYS40-GLH30"] == 0 #from prot

        # original
        assert out_dict["prot"]["LYS40-GLH30"] == .3
        # external
        assert out_dict["prot"]["GLU30-LYS40"] == 0  # from WT
        assert out_dict["prot"]["GLU30-LYS50"] == 0  # from WT
        assert out_dict["prot"]["LYS60-GLU30"] == 0  # from WT

        assert out_dict["prot"]["LYS60-GLU30"] == 0  # from K40A
        assert out_dict["prot"]["ALA40-GLU30"] == 0  # from K40A
        """

    def test_mutations(self):
        out_dict = str_and_dict.unify_freq_dicts(self.freq_dicts,
                                    key_separator=None,
                                    replacement_dict={"ALA40":"LYS40"})
        #original
        assert out_dict["WT"]["GLU30-LYS40"] == 1.0
        assert out_dict["WT"]["GLU30-LYS50"] == .5
        assert out_dict["WT"]["LYS60-GLU30"] == .1
        #external
        assert "ALA40-GLU30" not in out_dict["WT"].keys() #from K40A, will be now LYS40
        assert out_dict["WT"]["LYS40-GLU30"] == 0 #from K40A, was mutated back to LYS
        assert out_dict["WT"]["GLU30-LYS60"] == 0 #from K40A
        assert out_dict["WT"]["LYS40-GLH30"] == 0 #from prot

        # original
        assert out_dict["K40A"]["LYS40-GLU30"] == .7 #was mutated
        assert out_dict["K40A"]["GLU30-LYS60"] == .2
        # external
        assert out_dict["K40A"]["GLU30-LYS40"] == 0 #from WT
        assert out_dict["K40A"]["GLU30-LYS50"] == 0 #from WT
        assert out_dict["K40A"]["LYS40-GLH30"] == 0 #from prot

        # original
        assert out_dict["prot"]["LYS40-GLH30"] == .3
        # external
        assert out_dict["prot"]["GLU30-LYS40"] == 0  # from WT
        assert out_dict["prot"]["GLU30-LYS50"] == 0  # from WT
        assert out_dict["prot"]["LYS60-GLU30"] == 0  # from WT

        assert out_dict["prot"]["LYS60-GLU30"] == 0  # from K40A
        assert out_dict["prot"]["LYS40-GLU30"] == 0  # from K40A

    def test_reorder(self):
        out_dict = str_and_dict.unify_freq_dicts(self.freq_dicts,
                                    key_separator="-",
                                    )
        #original
        assert out_dict["WT"]["GLU30-LYS40"] == 1.0
        assert out_dict["WT"]["GLU30-LYS50"] == .5
        assert out_dict["WT"]["GLU30-LYS60"] == .1
        assert "LYS60-GLU30" not in out_dict["WT"].keys() #bc reordering
        #external
        assert out_dict["WT"]["ALA40-GLU30"] == 0 #from K40A
        assert out_dict["WT"]["GLH30-LYS40"] == 0 #from prot, reordered

        # original
        assert out_dict["K40A"]["ALA40-GLU30"] == .7
        assert out_dict["K40A"]["GLU30-LYS60"] == .2 #was reordered
        assert "LYS60-GLU30" not in out_dict["K40A"].keys()
        # external
        assert out_dict["K40A"]["GLU30-LYS40"] == 0 #from WT
        assert out_dict["K40A"]["GLU30-LYS50"] == 0 #from WT
        assert out_dict["K40A"]["GLH30-LYS40"] == 0 #from prot, reordered

        # original
        assert out_dict["prot"]["GLH30-LYS40"] == .3 #was reordered
        assert "LYS40-GLH30" not in out_dict["prot"].keys()

        # external
        assert out_dict["prot"]["GLU30-LYS40"] == 0  # from WT
        assert out_dict["prot"]["GLU30-LYS50"] == 0  # from WT
        assert out_dict["prot"]["GLU30-LYS60"] == 0  # from WT, reordered
        assert "LYS60-GLU30" not in out_dict["prot"].keys()

        #assert out_dict["prot"]["LYS60-GLU30"] == 0  # from K40A not anymore, was set to 0 already
        assert out_dict["prot"]["ALA40-GLU30"] == 0  # from K40A

    def test_per_residue(self):
        """
         self.freq_dicts = {"WT":   {"GLU30-LYS40":1.0,
                                    "GLU30-LYS50":.5,
                                    "LYS60-GLU30": .1},
                           "K40A": {"ALA40-GLU30":.7,
                                    "GLU30-LYS60": .2},
                           "prot": {"LYS40-GLH30": .3
                                    }}
        Returns
        -------

        """
        out_dict = str_and_dict.unify_freq_dicts(self.freq_dicts,
                                                 per_residue=True)

        test_dict = {"WT":   {"GLU30": 1.0 + .5 + .1,
                              "LYS40": 1.0 + 0,
                              "LYS50": .5,
                              "LYS60": .1,
                              "ALA40": 0,
                              "GLH30": 0},
                     "K40A": {"ALA40": .7,
                              "GLU30": .7 + .2,
                              "LYS60": .2,
                              "LYS40": 0,
                              "LYS50": 0,
                              "GLH30": 0},
                     "prot": {"GLU30":0,
                              "LYS40":.3,
                              "LYS50":0,
                              "LYS60":0,
                              "ALA40":0,
                              "GLH30":.3}
                     }


        self.assertDictEqual(out_dict["WT"], test_dict["WT"])
        self.assertDictEqual(out_dict["K40A"], test_dict["K40A"])
        self.assertDictEqual(out_dict["prot"], test_dict["prot"])

    def test_replacement_takes_place_before_reordering(self):
        ud = str_and_dict.unify_freq_dicts({"WT": {"A100-B200": 1.0},
                                            "MUT": {"C100-B200": 1.0}},
                                           replacement_dict={"C100": "A100"})

        self.assertDictEqual(ud, {'WT': {'A100-B200': 1.0}, 'MUT': {'A100-B200': 1.0}})

class Test_average_freq_dicts(unittest.TestCase):

    def setUp(self):
        self.freq_dicts = {"T300": {"GLU30-LYS40": 1.0,
                                    "LYS60-GLU30": .1},
                           "T310": {"GLU30-LYS40": .5,
                                    }
                           }



    def test_works(self):
        av_frqs = str_and_dict.average_freq_dict(self.freq_dicts)
        self.assertDictEqual(av_frqs,{"GLU30-LYS40":.75,"GLU30-LYS60":.05})

    def test_weights(self):
        av_frqs = str_and_dict.average_freq_dict(self.freq_dicts, weights={"T300":0, "T310":1})
        self.assertDictEqual(av_frqs, {"GLU30-LYS40": .5, "GLU30-LYS60": .0})

class Test_str_latex(unittest.TestCase):

    def setUp(self):
        self.greek_letters=["alpha", "beta", "gamma", "mu"]
        self.sub_supers = ["C_2", "C^2"]

    def test_works(self):
        np.testing.assert_array_equal(str_and_dict.replace4latex("There's an alpha and a beta here, also C_2"),
                                    "There's an $\\alpha$ and a $\\beta$ here, also $\\mathrm{{C}_{2}}$")

    def test_nothing_happens(self):
        istr = str_and_dict.replace4latex("There's an alpha and a beta here, also C_2")
        np.testing.assert_array_equal(str_and_dict.replace4latex(istr),
                                    istr)

class Test_auto_fragment_string(unittest.TestCase):

    def test_one_bad(self):
        assert str_and_dict.choose_options_descencing([None]) == ""

    def test_both_bad(self):
        assert str_and_dict.choose_options_descencing([None, None]) == ""

    def test_both_good(self):
        assert str_and_dict.choose_options_descencing(["3.50","fragA"]) == "3.50"

    def test_only_option(self):
        self.assertEqual(str_and_dict.choose_options_descencing([None, "fragA"]),"fragA")

    def test_only_better_option(self):
        assert str_and_dict.choose_options_descencing(["3.50",None]) == "3.50"

    def test_pick_best_label_just_works(self):
        assert (str_and_dict.choose_options_descencing(["Print this","Print this instead"]) == "Print this")

    def test_pick_best_label_exclude_works(self):
        assert(str_and_dict.choose_options_descencing([None,  "Print this instead"]) == "Print this instead")
        assert(str_and_dict.choose_options_descencing(["None","Print this instead"]) == "Print this instead")
        assert(str_and_dict.choose_options_descencing(["NA",  "Print this instead"]) == "Print this instead")
        assert(str_and_dict.choose_options_descencing(["na",  "Print this instead"]) == "Print this instead")

class Test_freq_file2dict(unittest.TestCase):

    def setUp(self):
        self.freqs = {"0-1":0.3, "0-2":.4}
        self.freqs_frags = {"0@TM1-1@TM2": 0.3, "0@TM1-2@TM2": .4}

    def test_works_w_ascii(self):
        with _TDir(suffix="_test_mdciao") as tempdir:
            tmpfile = path.join(tempdir, "frqfile.dat")
            with open(tmpfile,"w") as f:
                for key, val in self.freqs.items():
                    f.write("%f %s\n"%(val, key))

            freqsin = str_and_dict.freq_file2dict(tmpfile)
        assert freqsin["0-1"]==.3
        assert freqsin["0-2"]==.4
        assert len(freqsin)==2

    def test_works_w_excel_no_header(self):
        with _TDir(suffix="_test_mdciao") as tempdir:
            tmpfile = "freqfile.xlsx"
            tmpfile =  path.join(tempdir,tmpfile)
            with _XW(tmpfile) as writer:
                _DF.from_dict({"freq" : list(self.freqs.values()),
                               "label": list(self.freqs.keys())}, ).to_excel(writer,
                                                                             index=False,
                                                                             )

            freqsin = str_and_dict.freq_file2dict(tmpfile)
            self.assertEqual(freqsin["0-1"],.3)
            self.assertEqual(freqsin["0-2"],.4)
            assert len(freqsin) == 2

    def test_works_w_excel_w_header(self):
        with _TDir(suffix="_test_mdciao") as tempdir:
            tmpfile = "freqfile.xlsx"
            tmpfile =  path.join(tempdir,tmpfile)
            with _XW(tmpfile) as writer:
                _DF.from_dict({"freq" : list(self.freqs.values()),
                               "label": list(self.freqs.keys())}, ).to_excel(writer,
                                                                             index=False,
                                                                             startrow=1
                                                                             )

            freqsin = str_and_dict.freq_file2dict(tmpfile)
            self.assertEqual(freqsin["0-1"],.3)
            self.assertEqual(freqsin["0-2"],.4)
            assert len(freqsin) == 2

    def test_works_w_excel_w_header_defrag(self):
        with _TDir(suffix="_test_mdciao") as tempdir:
            tmpfile = "freqfile.xlsx"
            tmpfile =  path.join(tempdir,tmpfile)
            with _XW(tmpfile) as writer:
                _DF.from_dict({"freq" : list(self.freqs.values()),
                               "label": list(self.freqs.keys())}, ).to_excel(writer,
                                                                             index=False,
                                                                             startrow=1
                                                                             )

            freqsin = str_and_dict.freq_file2dict(tmpfile, defrag="@")
            self.assertEqual(freqsin["0-1"],.3)
            self.assertEqual(freqsin["0-2"],.4)
            assert len(freqsin) == 2


class Test_fnmatch_functions(unittest.TestCase):

    def setUp(self):
        self.dict = {"John Perez":[0,1],
                     "Johnathan Hernandez":[2,3],
                     "Guille Doe":[4,5],
                     "Johnny Doe":[6,7]}
        self.patterns = "John*,*Doe,-Johnny*"
        # pick all FN John, LN Doe but avoid Johnny

    def test_fnmatch_ex_works(self):

        names = self.dict.keys()
        filtered = str_and_dict.fnmatch_ex(self.patterns,names)
        self.assertEqual(filtered,["John Perez", "Johnathan Hernandez","Guille Doe"])

    def test_match_dict_by_patterns_works(self):
        matching_keys, matching_values = str_and_dict.match_dict_by_patterns(self.patterns,self.dict,
                                                      verbose=True)
        self.assertEqual(matching_keys,["John Perez", "Johnathan Hernandez","Guille Doe"])
        np.testing.assert_array_equal(matching_values,np.arange(6))
        # pick all FN John, LN Doe but avoid Johny

    def test_match_dict_by_patterns_empty(self):
        matching_keys, matching_values = str_and_dict.match_dict_by_patterns("Maria",
                                                                                   self.dict,
                                                      verbose=True)
        self.assertEqual(matching_keys,[])
        np.testing.assert_array_equal(matching_values,[])
        # pick all FN John, LN Doe but avoid Johny

class Test_aggregate_freq_dict_per_residue(unittest.TestCase):

    def test_works(self):
        indict = {"A-B":1.5, "B-C":10, "A-D":100, "D-C":1000}
        out_dict=str_and_dict.sum_dict_per_residue(indict,"-")
        np.testing.assert_equal(out_dict["A"], 1.5+100)
        np.testing.assert_equal(out_dict["B"], 1.5 + 10)
        np.testing.assert_equal(out_dict["C"], 10 + 1000)
        np.testing.assert_equal(out_dict["D"], 1000 + 100)

class Test_sort_dict(unittest.TestCase):

    def test_works(self):
        indict = {"A":0.5, "B":1.0, "C":.75}
        self.assertDictEqual(str_and_dict.sort_dict_by_asc_values(indict),               {"A": 0.5, "C": .75, "B": 1.0})
        self.assertDictEqual(str_and_dict.sort_dict_by_asc_values(indict, reverse=True), {"B": 1.0, "C": .75, "A": 0.5})


class Test_lexsort_ctc_labels(unittest.TestCase):

    def test_works(self):
        labels = ["ALA30@3.50-GLU50", "HIS28-GLU50", "ALA30-GLU20"]
        sorted_labels, order = str_and_dict.lexsort_ctc_labels(labels)
        self.assertListEqual(sorted_labels, ["HIS28-GLU50", "ALA30-GLU20", "ALA30@3.50-GLU50"])
        self.assertListEqual(order.tolist(), [1, 2, 0])

    def test_works_one_residue(self):
        labels = ["GLU50", "ALA30", "GLU20"]
        sorted_labels, order = str_and_dict.lexsort_ctc_labels(labels)
        self.assertListEqual(sorted_labels, ["GLU20", "ALA30", "GLU50"])
        self.assertListEqual(order.tolist(), [2, 1, 0])
    def test_works_reverse(self):
        labels = ["ALA30@3.50-GLU50", "HIS28-GLU50", "ALA30-GLU20"]
        sorted_labels, order = str_and_dict.lexsort_ctc_labels(labels, reverse=True)
        self.assertListEqual(sorted_labels, ["HIS28-GLU50", "ALA30-GLU20", "ALA30@3.50-GLU50"][::-1])
        self.assertListEqual(order.tolist(), [1, 2, 0][::-1])

    def test_works_second_column_first(self):
        labels = ["ALA30@3.50-GLU50", "HIS28-GLU50", "ALA30-GLU20"]
        sorted_labels, order = str_and_dict.lexsort_ctc_labels(labels, columns=[1, 0])
        self.assertListEqual(sorted_labels, ["ALA30-GLU20", "HIS28-GLU50", "ALA30@3.50-GLU50"])
        self.assertListEqual(order.tolist(), [2, 1, 0])

    def test_works_second_column_first_reverse(self):
        labels = ["ALA30@3.50-GLU50", "HIS28-GLU50", "ALA30-GLU20"]
        sorted_labels, order = str_and_dict.lexsort_ctc_labels(labels, columns=[1, 0], reverse=True)
        self.assertListEqual(sorted_labels, ["ALA30@3.50-GLU50", "HIS28-GLU50", "ALA30-GLU20"])
        self.assertListEqual(order.tolist(), [0, 1, 2])

class Test_label2componentsdict(unittest.TestCase):

    def setUp(self):
        self.w = {
            1: "r1",
            2: "r1@f1",
            3: "r1@f1-r2",
            4: "r1@f1-r2",
            5: "r1@f1-r2@f2",
            6: "r1@f1-1-r2@f2",
            7: "r1@f1-r2@f2-2",
            8: "r1@f1-1-r2@f2-2",
            9: "r1-r2",
           10: "r1-r2@f2",
           11: "r1-r2@f2-1",
           12: "r1@f1-1-1-r2@f2-2-2"
        }

    def test_1(self):
        self.assertDictEqual({"res1": "r1"},
                             str_and_dict._label2componentsdict(self.w[1]),
                             self.w[1])

    def test_2(self):
        self.assertDictEqual({"res1": "r1",
                              "frag1": "f1"},
                             str_and_dict._label2componentsdict(self.w[2]),
                             self.w[2])

    def test_3(self):
        self.assertDictEqual({"res1": "r1",
                              "frag1": "f1",
                              "res2": "r2"
                              },
                             str_and_dict._label2componentsdict(self.w[3]),
                             self.w[3])

    def test_4_assume_False(self):
        self.assertDictEqual({"res1": "r1",
                              "frag1": "f1-r2",
                              },
                             str_and_dict._label2componentsdict(self.w[4],
                                                                assume_ctc_label=False),
                             self.w[4])

    def test_5(self):
        self.assertDictEqual({"res1": "r1",
                              "frag1": "f1",
                              "res2": "r2",
                              "frag2": "f2"},
                             str_and_dict._label2componentsdict(self.w[5]),
                             self.w[5])


    def test_6(self):
        self.assertDictEqual({"res1": "r1",
                              "frag1": "f1-1",
                              "res2": "r2",
                              "frag2": "f2"},
                             str_and_dict._label2componentsdict(self.w[6]),
                             self.w[6])

    def test_7(self):
        self.assertDictEqual({"res1": "r1",
                              "frag1": "f1",
                              "res2": "r2",
                              "frag2": "f2-2"},
                             str_and_dict._label2componentsdict(self.w[7]),
                             self.w[7])

    def test_8(self):
        self.assertDictEqual({"res1": "r1",
                              "frag1": "f1-1",
                              "res2": "r2",
                              "frag2": "f2-2"},
                             str_and_dict._label2componentsdict(self.w[8]),
                             self.w[8])

    def test_9(self):
        self.assertDictEqual({"res1": "r1",
                              "res2": "r2",
                              },
                             str_and_dict._label2componentsdict(self.w[9]),
                             self.w[9])

    def test_10(self):
        self.assertDictEqual({"res1": "r1",
                              "res2": "r2",
                              "frag2":"f2",
                              },
                             str_and_dict._label2componentsdict(self.w[10]),
                             self.w[10])

    def test_11(self):
        self.assertDictEqual({"res1": "r1",
                              "res2": "r2",
                              "frag2": "f2-1",
                              },
                             str_and_dict._label2componentsdict(self.w[11]),
                             self.w[11])

    def test_12(self):
        self.assertDictEqual({"res1": "r1",
                              "res2": "r2",
                              "frag1": "f1-1-1",
                              "frag2": "f2-2-2",
                              },
                             str_and_dict._label2componentsdict(self.w[12]),
                             self.w[11])

    def test_dont_split(self):
        label = "K417@RBD-Cl-1115@ions"
        # wrong behavior
        self.assertDictEqual({'res1': 'K417',
                              'frag1': 'RBD-Cl',  # wrong frag1
                              'res2': '1115',  # wrong res2
                              'frag2': 'ions'
                              },
                             str_and_dict._label2componentsdict(label),
                             label)
        # right behavior
        self.assertDictEqual({'res1': 'K417',
                              'frag1': 'RBD',  # right frag1
                              'res2': 'Cl-1115',  # right res2
                              'frag2': 'ions'
                              },
                             str_and_dict._label2componentsdict(label, dont_split=["Cl-1115"]),
                             label)

class Test_splitlabel(unittest.TestCase):

    def setUp(self):
        self.w = {
            3: "r1@f1-r2",
            5: "r1@f1-r2@f2",
            6: "r1@f1-1-r2@f2",
            7: "r1@f1-r2@f2-2",
            8: "r1@f1-1-r2@f2-2",
            9: "r1-r2",
           10: "r1-r2@f2",
           11: "r1-r2@f2-1",
           12: "r1@f1-1-1-r2@f2-2-2"

        }

    def test_3(self):
        self.assertListEqual(["r1@f1","r2"], str_and_dict.splitlabel(self.w[3]))

    def test_5(self):
        self.assertListEqual(["r1@f1","r2@f2"], str_and_dict.splitlabel(self.w[5]))

    def test_6(self):
        self.assertListEqual(["r1@f1-1","r2@f2"], str_and_dict.splitlabel(self.w[6]))

    def test_7(self):
        self.assertListEqual(["r1@f1","r2@f2-2"], str_and_dict.splitlabel(self.w[7]))

    def test_8(self):
        self.assertListEqual(["r1@f1-1","r2@f2-2"], str_and_dict.splitlabel(self.w[8]))

    def test_9(self):
        self.assertListEqual(["r1","r2"], str_and_dict.splitlabel(self.w[9]))

    def test_10(self):
        self.assertListEqual(["r1","r2@f2"], str_and_dict.splitlabel(self.w[10]))

    def test_11(self):
        self.assertListEqual(["r1","r2@f2-1"], str_and_dict.splitlabel(self.w[11]))

    def test_12(self):
        self.assertListEqual(["r1@f1-1-1", "r2@f2-2-2"], str_and_dict.splitlabel(self.w[12]))

class Test_defrag(unittest.TestCase):

    def test_works(self):
        label = 'res1@frag1-res2@frag2'
        np.testing.assert_equal("res1-res2", str_and_dict.defrag_key(label))

    def test_works_one_missing_frag(self):
        label = 'res1@frag1-res2    '
        np.testing.assert_equal("res1-res2", str_and_dict.defrag_key(label))

    def test_works_no_frags(self):
        label = 'res1-res2    '
        np.testing.assert_equal("res1-res2", str_and_dict.defrag_key(label))

    def test_works_just_one(self):
        label = 'res1@frag1    '
        np.testing.assert_equal("res1", str_and_dict.defrag_key(label))

class Test_replace_regex_special_chars():
    def test_works(self):
        word ="[]()^"
        assert str_and_dict._replace_regex_special_chars(word,"!!!!!")

class Test_latex_mathmode(unittest.TestCase):

    def test_works(self):
        self.assertEqual("$\\mathrm{There's an }\\alpha\\mathrm{ and a }\\beta\\mathrm{ here, also C_200}$",
                         str_and_dict.latex_mathmode("There's an alpha and a beta here, also C_200"))

class Test_latex_superscript_one_fragment(unittest.TestCase):

    def test_works(self):
        self.assertEqual(str_and_dict._latex_superscript_one_fragment("GLU30@beta_2AR"),
                         "GLU30$^{\\beta\\mathrm{_2AR}}$")
    def test_no_fragment(self):
        self.assertEqual(str_and_dict._latex_superscript_one_fragment("GLU30"),"GLU30")
if __name__ == '__main__':
    unittest.main()

class Test_FilenameGenerator(unittest.TestCase):

    def test_just_runs(self):
        fn = str_and_dict.FilenameGenerator("beta2 Gs",3.5,"project","png", "dat",150,"ps")
        self.assertEqual(fn.fullpath_overall_fig, "project/beta2_Gs.overall@3.5_Ang.png")
        self.assertEqual(fn.fullpath_overall_excel, "project/beta2_Gs.overall@3.5_Ang.xlsx")
        self.assertEqual(fn.fullpath_overall_dat, "project/beta2_Gs.overall@3.5_Ang.dat")
        self.assertEqual(fn.fullpath_pdb, "project/beta2_Gs.overall@3.5_Ang.as_bfactors.pdb")
        self.assertEqual(fn.fullpath_matrix, "project/beta2_Gs.matrix@3.5_Ang.png")
        self.assertEqual(fn.fullpath_flare_vec, "project/beta2_Gs.flare@3.5_Ang.pdf")
        self.assertEqual(fn.table_ext,"dat")
        self.assertEqual(fn.graphic_dpi,150)
        self.assertEqual(fn.t_unit,"ps")
        self.assertEqual(fn.fname_per_site_table("NPY"),'project/beta2_Gs.NPY@3.5_Ang.dat')
        self.assertEqual(fn.fname_timetrace_fig("traj1"),'beta2_Gs.traj1.time_trace@3.5_Ang.png')

    def test_table_ext_None_raises(self):
        with self.assertRaises(ValueError):
            fn = str_and_dict.FilenameGenerator("beta2 Gs",3.5,"project","png",None,150,"ps")

    def test_keeps_svg(self):
        fn = str_and_dict.FilenameGenerator("beta2 Gs",3.5,"project","svg", "dat",150,"ps")
        self.assertEqual(fn.fullpath_flare_vec, "project/beta2_Gs.flare@3.5_Ang.svg")

class Test_print_wrap(unittest.TestCase):

    def test_just_prints(self):
        string = "AAAABBBB"
        b = _io.StringIO()
        with _contextlib.redirect_stdout(b):
            str_and_dict.print_wrap(string, 4)
        self.assertEqual(b.getvalue(), "AAAA\nBBBB\n")
    def test_just_returns(self):
        string = "AAAABBBB"
        wrapped = str_and_dict.print_wrap(string,4, just_return_string=True)
        self.assertEqual(wrapped, "AAAA\nBBBB")

class Test_intblocks_in_str(unittest.TestCase):

    def test_just_works(self):
        self.assertListEqual(str_and_dict.intblocks_in_str("GLU30@3.50-GDP396@frag1"), [30,3,50,396,1])
    def test_fails(self):
        with self.assertRaises(ValueError):
            str_and_dict.intblocks_in_str("GLU")











import os
from shutil import rmtree
import unittest
from mdciao.examples import examples, filenames
from mdciao import contacts
from tempfile import mkdtemp
import contextlib
import numpy as _np


@contextlib.contextmanager
def remember_cwd():
    curdir = os.getcwd()
    try:
        yield
    finally:
        os.chdir(curdir)


class Test_ExamplesCLTs(unittest.TestCase):

    # It's already a test!
    def setUp(self):
        self.xCLTs = examples.ExamplesCLTs(
            test=True
        )
        self.tmpdir = mkdtemp(suffix="mdciao_tests")
        print(self.tmpdir)

        for fn in [filenames.traj_xtc,
                   filenames.top_pdb,
                   filenames.adrb2_human_xlsx,
                   filenames.gnas2_human_xlsx,
                   filenames.tip_json
                   ]:
            os.symlink(fn, os.path.join(self.tmpdir, os.path.basename(fn)))

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_mdc_sites(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_sites",
                                # show=True
                                )
        assert _np.unique([iCP.returncode for iCP in CP]) == 0

    def test_mdc_neighborhood(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_neighborhoods",
                                # show=True
                                )
        assert _np.unique([iCP.returncode for iCP in CP]) == 0

    def test_mdc_interface(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_interface",
                                # show=True
                                )
        assert _np.unique([iCP.returncode for iCP in CP]) == 0

    def test_mdc_fragments(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_fragments",
                                # show=True
                                )
        assert _np.unique([iCP.returncode for iCP in CP]) == 0

    def test_mdc_BW(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_GPCR_overview",
                                # show=True
                                )
        assert _np.unique([iCP.returncode for iCP in CP]) == 0, CP

    def test_mdc_CGN(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_CGN_overview",
                                # show=True
                                )
        assert _np.unique([iCP.returncode for iCP in CP]) == 0

    def test_mdc_pdb(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_pdb",
                                # show=True
                                )
        assert _np.unique([iCP.returncode for iCP in CP]) == 0

    def test_mdc_compare(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_compare")
            assert _np.unique([iCP.returncode for iCP in CP]) == 0

    def test_mdc_residues(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_residues")
            assert _np.unique([iCP.returncode for iCP in CP]) == 0

    def test_mdc_notebooks(self):
        with remember_cwd():
            os.chdir(self.tmpdir)
            CP = self.xCLTs.run("mdc_notebooks")
            assert _np.unique([iCP.returncode for iCP in CP]) == 0


if __name__ == '__main__':
    unittest.main()



import unittest

from tempfile import TemporaryDirectory as _TD
import os
import contextlib
@contextlib.contextmanager
def remember_cwd():
    curdir = os.getcwd()
    try:
        yield
    finally:
        os.chdir(curdir)


from mdciao import pdb
class Test_pdb2ref(unittest.TestCase):

    def test_works(self):
        cite = pdb.pdb2ref("3SN6")
        assert isinstance(cite,dict)
        assert all([key in cite.keys() for key in ["title", "rcsb_authors", "rcsb_journal_abbrev","year"]])

    def test_work_wrong_PDB(self):
        cite = pdb.pdb2ref("0SN6")
        assert cite is None

    def test_work_with_to_be_published(self):
        cite = pdb.pdb2ref("7dua")
        assert isinstance(cite,dict)
    def test_work_wo_doi(self):
        cite = pdb.pdb2ref("6vg3")
        assert isinstance(cite, dict)

class Test_url2json(unittest.TestCase):

    def test_works(self):
        idict = pdb.pdb._url2json("https://data.rcsb.org/rest/v1/core/entry/3SN6",
                                                    5,True)
        assert isinstance(idict,dict)

    def test_work_url_works_json_is_404(self):
        idict = pdb.pdb._url2json("https://data.rcsb.org/rest/v1/core/entry/13SN6",
                                                    5,True)
        assert idict["status"] == 404

    def test_work_url_works_cant_json(self):
        idict = pdb.pdb._url2json("https://rcsb.org",5, True)
        assert isinstance(idict,ValueError)


class Test_pdb2traj(unittest.TestCase):
    def test_works(self):
        with _TD(suffix="_test_mdciao") as tmpdir:
            pdb.pdb2traj("3SN6",os.path.join(tmpdir,"3SN6.pdb"))

    def test_wrong_url(self):
        pdb.pdb2traj("3SNXX")



r"""
Guess and manipulate fragments, i.e., sub-regions
of molecular topologies. A fragment is just an iterable
of integers representing the residue indices belonging to it.

.. currentmodule:: mdciao.fragments

.. autosummary::
   :toctree: generated/
   :nosignatures:

    get_fragments
    fragment_slice
    print_fragments
    print_frag
    overview
    match_fragments
    assign_fragments


"""
from .fragments import *
from .fragments import _get_fragments_by_jumps_in_sequence


r"""
Functions related to Center-Of-Mass (COM) computations

.. autosummary::
   :nosignatures:
   :toctree: generated/


"""
import numpy as _np
import mdtraj as _md
from tqdm import tqdm as _tqdm
from mdtraj.geometry.distance import compute_distances_core as _compute_distances_core

def geom2COMdist(geom, residue_pairs, subtract_max_radii=False, low_mem=True,
                 periodic=True, per_residue_unwrap=True) -> _np.ndarray:
    r"""
    Returns the time-trace of the distances between pairs of residues' centers-of-mass (COM)

    The option `subtract_max_radii` can be used to produce a time-dependent lower bound
    on the distance between any atoms of each residue pair, i.e. a lower bound
    on the pairwise-residue distance. This lower bound can then be used to discard
    any contact between some pairs residues in any frame of `geom`,
    without having to compute all pairwise atom distances between the `residue_pairs`.

    Please see below the `periodic` and `per_residue_unwrap` for how periodic-boundary-conditions
    affect this calculation.

    Parameters
    ----------
    geom: :obj:`mdtraj.Trajectory`
    residue_pairs: iterable of pairs of integers
        zero-indexed pairs of residues
    subtract_max_radii : bool, default is False
        Subtract the sum of maximum radii (as computed by
        :obj:`~mdciao.utils.COM.geom2max_residue_radius`)
        of both residues of each residue pair from
        the residue-residue COM distance, effectively producing
        a lower bound for the residue-residue mindist.
        Please note that this option can produce negative values,
        meaning the (very normal situation) of two residue COMs
        being closer to each than the sum their max-radii.
    low_mem : bool, default is True
        Try to use less memory by using each residue's
        maximum radius for all frames (instead of a frame-dependent value)
        when subtracting pairs of radii from the COM distances.
        This results in an even lower value for lower bound
        of the COM-distances, which is itself still a lower-bond.
        For the purposes of cutoff-thresholding, this "overshooting"
        has little consequence, see the note below for a benchmark case.
    periodic : bool, default is True
        Compute COM distances under the minimum image convention.
        Please see the warning in :obj:`~mdciao.utils.COM.COMs_xyz`
        wrt residues being split across periodic boundaries into
        adjacent periodic images and the impact on COM computation.
    per_residue_unwrap : bool, default is True
        Unwrap residues individually for the purpose of
        computing the residue COM. This is to deal with
        the warning in :obj:`~mdciao.utils.COM.geom2COMxyz`.
        The original coordinates in `geom` remain unchanged.
        Since periodic boundary conditions are necessarily
        used when unwrapping, and the unwrapping doesn't make
        the entire `geom` whole, but only residues, the method
        will fail if `per_residue_unwrap` is True but `periodic` is False,
        alerting the user of what they are trying to do. See the
        note below for information on the overhead induced
        by the `per_residue_unwrap`.

    Note
    ----
        In a benchmark case of 6K frames and ~100K `residue_pairs`,
        after using `subtract_max_radii=True`, we checked for `residue_pairs`
        with a lower-bound for their residue-residue distance
        smaller than a given cutoff, deeming them potential contacts.
        As expected, thresholding the values obtained with
        `low_mem=True` resulted in longer lists of potential
        contacts (between 20%-30% longer depending on the cutoff),
        and accordingly longer computation times when calling
        :obj:`mdtraj.compute_contacts` on the kept `residue_pairs`. However,
        those "longer" lists (and times) are not significant at all when expressed as a
        percentage of the initial `residue_pairs`:

        * Using a cutoff of 1 Angstrom when thresholding the mindist
          values obtained with `low_mem=False` discarded 99.46% of the original `residue_pairs`,
          whereas using the `low_mem=True` values meant discarding "only" 99.29%.
        * Using a cutoff of 6 Angstrom when thresholding the mindist
          values obtained with `low_mem=False` discarded 98.3% of the original `residue_pairs`,
          whereas using the `low_mem=True` values meant discarding "only" 98%.

        So there is still a huge reduction in the number of potential contact-pairs
        when thresholding, from ca 100K residue pairs to between 500 and 2K.
        For this benchmark system, `low_mem=True` used ~10GBs of memory
        and `low_mem=False` used ~15GB. There's a table included
        as a comment in the source code of the method showing the benchmark numbers.

    Note
    ----
        Just how much overhead the per-residue unwrapping adds is hard to know a-priori.
        In our benchmark system, it can be between 5% or 25% of the overall computation time.
        It simply depends on how many residues are split across PBCs in how many frames,
        which depends highly on what post-processing steps the trajectory has undergone.
        * The 5% overhead is when no residues have to be unwrapped because the pre-processing
         has taken care of this, e.g. with `gmx trjconv -pbc whole`. The 5% is just the
         time used to check whether the residues need to be unwrapped or not.
        * The 25% overhead is for a plausible worst-case, when the split across PBCs
         affects a high number of residues. The system is effectively halved at
         the plane where most residues are affected simultaneously by the split.
        * For the highly-constructed, and physically inplausible
         situation where **all** are residues need to be unwrapped, the overhead is 70%.
         This is very unlikely, because it implies that **all** residues have some atoms
         just a few Angstrom away from the boundary

        In essence, you can always leave `per_residue_unwrap=True` on unless significant slowdown
        is noticed. Even better, if you notice a significant slowdown, why not pre-process
        your trajectory to be whole and centered in the box (which is sometimes called
        unwrapping), and then there's no need to use `per_residue_unwrap`. Note, even in that
        case a decision needs to be made whether to use PBCs when computing distances
        between residue COMs


    Returns
    -------
    COMs_array : np.ndarray of shape(geom.n_frames, len(res_pairs))
        contains the time-traces of the distance between residue COMs,
        optionally minus the sum of each pair of residue radii.
        Please note that this option can produce negative values,
        meaning the residue-spheres actually intersect at some point.
    """

    residue_pairs = _np.array(residue_pairs)
    residue_idxs_unique, pair_map = _np.unique(residue_pairs, return_inverse=True)
    pair_map = pair_map.reshape(len(residue_pairs),2)

    if per_residue_unwrap:
        assert periodic, ValueError("Cannot unwrap residues if 'periodic' is set to False.")
        # Per-residue per-frame unwraping
        unwrapped_residue_geom = _per_residue_unwrapping(geom,residue_idxs=residue_idxs_unique)

    else:
        unwrapped_residue_geom = geom
    # This would be worth migrating to mdanalysis
    # https://docs.mdanalysis.org/1.0.1/documentation_pages/core/groups.html#MDAnalysis.core.groups.ResidueGroup.center
    COMs_xyz = geom2COMxyz(unwrapped_residue_geom, residue_idxs=residue_idxs_unique)

    COM_dists_t = _compute_distances_core(COMs_xyz,
                                          pair_map,
                                          unitcell_vectors=geom.unitcell_vectors,
                                          periodic=periodic,
                                          )

    if subtract_max_radii:
        if low_mem:
            res_max_radius = geom2max_residue_radius(unwrapped_residue_geom, residue_idxs_unique, res_COMs=COMs_xyz).max(0)
            max_radius_pairs = res_max_radius[pair_map].sum(1)
        else:
            res_max_radius = geom2max_residue_radius(unwrapped_residue_geom, residue_idxs_unique, res_COMs=COMs_xyz)
            max_radius_pairs = res_max_radius[:, pair_map].sum(axis=-1)

        # TODO update memory numbers
        # Low mem vs high mem. 6000 frames, 106499 atoms
        # peak memory: 22178.34 MiB, increment: 10344.77 MiB
        # peak memory: 27577.62 MiB, increment: 15739.57 MiB <-high mem
        #|    |   cutoff / nm |   #pairs lo-mem |   #pairs hi-mem |   t lo-mem / s |   t hi-mem / s |    #pairs lo-mem / |    #pairs hi-mem / |    #pairs lo-mem / |
        #|    |               |                 |                 |                |                |    % residue_pairs |    % residue_pairs |    % #pairs hi-mem |
        #|---:|--------------:|----------------:|----------------:|---------------:|---------------:|-------------------:|-------------------:|-------------------:|
        #|  1 |           0.1 |             751 |             573 |           4.27 |           3.41 |               0.71 |               0.54 |             131.06 |
        #|  2 |           0.2 |             963 |             770 |           5.21 |           4.31 |               0.9  |               0.72 |             125.06 |
        #|  3 |           0.3 |            1198 |             979 |           6.33 |           5.21 |               1.12 |               0.92 |             122.37 |
        #|  4 |           0.4 |            1479 |            1219 |           7.59 |           6.3  |               1.39 |               1.14 |             121.33 |
        #|  5 |           0.5 |            1780 |            1496 |           9.05 |           7.43 |               1.67 |               1.4  |             118.98 |
        #|  6 |           0.6 |            2117 |            1796 |          10.4  |           8.98 |               1.99 |               1.69 |             117.87 |
        COM_dists_t -= max_radius_pairs

    return COM_dists_t

def geom2COMxyz(igeom, residue_idxs=None):
    r"""
    Returns the time-trace of per-residue center-of-masses (COMs)

    Warning
    -------
    In some cases, molecules may have residues split across the periodic boundaries,
    i.e. all atoms are within the periodic unit cell, but atoms of the same
    residue are scattered in different places inside the cell, near the walls.
    Then, the residue COM might fall between these fragments, far away from any
    of the residues' atoms. In that case, the residue COM is not a useful measure for
    the residue's approximate position.

    Parameters
    ----------
    igeom : :obj:`mdtraj.Trajectory`

    residue_idxs : iterable, default is None
        Residues for which the center of mass will be computed. Default
        is to compute all residues.

    Returns
    -------
    COMs : numpy.ndarray of shape (igeom.n_frames, len(residue_idxs),3)

    """

    if residue_idxs is None:
        residue_idxs=_np.arange(igeom.top.n_residues)
    masses = [_np.hstack([aa.element.mass for aa in rr.atoms]) for rr in
              igeom.top.residues]
    COMs_res_time_coords = _np.array([_np.average(igeom.xyz[:, [aa.index for aa in igeom.top.residue(index).atoms], :],
                                                  axis=1, weights=masses[index])
                                      for index in residue_idxs])
    COMs_time_res_coords = _np.swapaxes(_np.array(COMs_res_time_coords), 0, 1)
    return COMs_time_res_coords

def geom2max_residue_radius(geom, residue_idxs=None, res_COMs=None) -> _np.ndarray:
    r"""
    Per-residue maximum radius, i.e. the maximum distance between any atom of the residue and the residue's center of mass

    Warning
    -------
    No periodic-boundary-conditions are taken into account, i.e.
    residues are assumed "whole" or "unwrapped", and then the mass-weighted
    average of a residues's atoms' cartesian coordinates are computed. If your residues
    are not "whole", i.e. atoms o the same residue, then the residue COM
    might be meaningless (check the warning in :obj:`~mdciao.utils.COM.geom2COMxyz`)
    and so will be the maximum residue radius.

    Parameters
    ----------
    geom : :obj:`mdtraj.Trajectory`
    residue_idxs : iterable of ints, default is None
        The indices of the residues for which
        the residue radius will be computed.
        If None, all residues will be considered.
        If `residue_idxs` is None, then `res_COM` has to
        be None as well, else you would be
        providing `res_COMs` any information
        on what residues they belong to.
    res_COMs : np.ndarray, default is None
        The time-traces of the xyz coordinates of
        the residue COMs. Has to have
        shape(geom.n_frames, len(residue_idxs), 3).
        It will be computed on the fly if None is provided.

    Returns
    -------
    r : numpy.ndarray
        Shape (igeom.n_frames, len(residue_idxs))

    """
    if residue_idxs is None:
        residue_idxs = _np.arange(geom.n_residues)
        if res_COMs is not None:
            raise ValueError("If 'residue_idxs' is None, then 'res_COMs' has to be None as well.")

    if res_COMs is None:
        res_COMs = geom2COMxyz(geom, residue_idxs=residue_idxs)
    else:
        assert res_COMs.shape[0] == geom.n_frames
        assert res_COMs.shape[1] == len(residue_idxs)
    atom_idxs = [[aa.index for aa in geom.top.residue(rr).atoms] for rr in residue_idxs]

    r = [_np.linalg.norm(geom.xyz[:, idxs, :] - res_COMs[:, _np.newaxis, ii], axis=-1).max(axis=1) for ii, idxs in
         enumerate(atom_idxs)]

    r = _np.vstack(r).T

    return r


def _per_residue_unwrapping(traj,
                            residue_idxs=None,
                            max_res_radii_t=None, progressbar=False, inplace=False,
                            unwrap_when_smaller_than=.4) -> _md.Trajectory:
    r"""

    Translate (if needed) atoms of a residue to a different periodic image to make the residue "whole"

    Sometimes this is referred to as "unwrapping".
    To decide whether a given residue needs unwrapping
    in a given frame, the method checks if any of the residue's
    atoms are further away from the residue COM than a half the
    box dimensions. In theory, one would check the distance
    per-dimension (distance in x > .5 * box_x,  in y > .5 box_y etc),
    as in the minimum image convention, also taking into account
    frame-dependent box dimensions. Practically, it's faster to
    compare euclidean distance in 3D to the COM (i.e. the maximum
    residue radius) against the smallest box length for all frames.
    This might "overshoot" and try to unwrap a residue that
    doesn't need unwrapping, with no other effect than
    leaving that residues intact.

    Ensuring unbroken (sometimes called "whole") residues,
    even if that means having atoms of the same residue spread
    across adjacent periodic images allows for the
    computation of per-residue center-of-mass (COM) that
    wont place COMs "in the middle the box" when a residue
    has been split across the PBCs and "wrapped" into the box.

    Note
    ----
    This method only makes **residues** whole, but not necessarily the
    entire molecule itself, i.e. it is not :obj:`mdtraj.Trajectory.image_molecule`.
    The molecule itself, i.e. the bonds between residues
    might still be split across PBCs after running this method.
    What this method guarantees is that residue COMs
    have meaning and are placed at the center of the group of atoms that form
    the "whole" residue. This is very useful as a preprocessing step to compute
    lower bounds for residue distances under PBCs, because residue positions
    and maximum residue radii are meaningful even if the COMs are spread across
    different periodic images.

    Parameters
    ----------
    traj : :obj:`~mdtraj.Trajectory`
    residue_idxs : iterable of ints, default is None
        Apply the method only to these residues. Default
        is to apply it to all.
    max_res_radii_t : 2D np.ndarray, default is None
        Optionally, pass here the value of
        the residue radii before any unwrapping takes place,
        else it will be computed on the fly.
        Shape is (traj.n_frames, traj.n_residues).
    progressbar : bool, default is False
        Show progressbar.
    inplace : bool, default is False
        Change the coordinates of `traj`
        inplace. Default (False) is to
        return a new object and leave the
        original `traj`  untouched.
    unwrap_when_smaller_than : float, default is .4
        Unwrap when the maximum residue radius
        is larger than the minimum box length times
        this factor. Flagging residues
        that don't need unwrapping for unwrapping
        doesn't have any effect

    Returns
    -------
    out : :obj:`~mdtraj aj.Trajectory`
        A new trajectory with unwrapped
        residues or, if `inplace` was True,
        an updated `traj` with the unwrapped residues.
    """
    if residue_idxs is None:
        residue_idxs = _np.arange(traj.n_residues)
    if max_res_radii_t is None:
        max_res_radii_t = geom2max_residue_radius(traj, residue_idxs=residue_idxs)
    else:
        assert max_res_radii_t.shape ==(traj.n_frames, len(residue_idxs))
    broken_res_bool = max_res_radii_t > (traj.unitcell_lengths.min() * unwrap_when_smaller_than)
    assert broken_res_bool.shape == (traj.n_frames, len(residue_idxs)) #making extra sure nothing goes wrong here
    if inplace:
        outtraj = traj
    else:
        outtraj = _md.Trajectory(_np.copy(traj._xyz), traj.top, time=traj.time, unitcell_angles=traj.unitcell_angles, unitcell_lengths=traj.unitcell_lengths)

    if broken_res_bool.any():
        PB = _tqdm(total = broken_res_bool.sum(), disable=not progressbar)
        for res_idx, res_bool in zip(residue_idxs, broken_res_bool.T):
            time_frames = _np.flatnonzero(res_bool)
            if len(time_frames) > 0:
                atoms = [aa.index for aa in traj.top.residue(res_idx).atoms]
                whole_res_xyz_t = _unwrap(traj.xyz[time_frames][:, atoms, :], traj.unitcell_lengths[time_frames, :])
                #print(res_idx, time_frames.shape)
                for ff, ixyz in zip(time_frames, whole_res_xyz_t):
                    outtraj._xyz[ff, atoms, :] = ixyz
                PB.update(len(time_frames))
    return outtraj

_translations = _np.vstack(_np.vstack([[[(kk, jj, ii) for ii in range(-1,2)] for jj in range(-1,2)] for kk in range(-1,2)]))

def _unwrap(xyz_t, unitcell_lengths):
    r"""

    For a trajectory of xyz coordinates, typically from atoms
    belonging to the same residue, apply, for each atom,
    the periodic translation that places that atom in the
    same periodic image as the rest of the atoms.

    In 1D, a residue with 5 atoms is split across the
    boundary:
        |cde-----------ab|
    This method picks the atom closest to the center of the box (e)
    and finds, for the other atoms, which one of the periodic
    translations, -1 * box_x, 0 * box_x, or +1 * box_x,
    places each atom a,b,c,d the closet to e, regardless
    of splitting the residues across the PBC
    In this case, For c,d, the translation is "no translation",
    i.e. 0 * box_x, whereas for a and b
    it will be -1 * box_x, yielding
      ab|cde-------------|
    Once the translation has been decided, it is applied
    on the coordinates and the coordinates returned.

    For each frame, the frame-dependent `unitcell_lengths`
    PBCs are taken into account.

    Parameters
    ----------
    xyz_t : 3D np.ndarray
        (n_frames, n_atoms, 3)
    unitcell_lengths : 2D np.ndarray
        (n_frames, 3)

    Returns
    -------
    res_xyz_t : 3D np.ndarray
        (n_frames, n_atoms, 3)
    """
    translations_vec = _translations * unitcell_lengths[:, _np.newaxis, :]
    d2center = _np.linalg.norm(xyz_t - unitcell_lengths[:, _np.newaxis, :] / 2, axis=2)
    most_centered_atom = d2center.argmin(axis=1)
    xyz_best_centered_atom = _np.vstack([xyz_t[ii][jj] for ii, jj in enumerate(most_centered_atom)])
    xyz_t -= xyz_best_centered_atom[:, _np.newaxis, :]
    # Doing these two operations with integers hardly speeds up the computation
    translated_coords = xyz_t[:, :, _np.newaxis, :] + translations_vec[:, _np.newaxis, :, :]
    d2tran = _np.linalg.norm(translated_coords, axis=-1)
    # Since we're only interested in the argmin we could've just computed
    # the sum of the abs(components), which respects the order of values,
    # but there's hardly any speedup
    best_translation_idx_per_frame_per_atom = d2tran.argmin(axis=-1)

    final_trans = []
    # This loop could/should be optimized to vector operations but currently
    # it's only %5 of compute time (cf the translated_coords + d2tran which is 70%)
    for ii, row in enumerate(best_translation_idx_per_frame_per_atom):
        final_trans.append(translations_vec[ii,row])
    final_trans = _np.array(final_trans)
    xyz_t += final_trans + xyz_best_centered_atom[:, _np.newaxis, :]
    return xyz_t

